#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import annotations

import argparse
from unittest import mock

import testtools

from ironicclient.osc import plugin
from ironicclient.tests.unit.osc import fakes
from ironicclient.v1 import client


class MakeClientTest(testtools.TestCase):

    @mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=False)
    @mock.patch.object(client, 'Client', autospec=True)
    def test_make_client_explicit_version(
            self, mock_client: mock.Mock) -> None:
        instance = fakes.FakeClientManager()
        instance.get_endpoint_for_service_type = mock.Mock(
            return_value='endpoint')
        plugin.make_client(instance)
        mock_client.assert_called_once_with(
            os_ironic_api_version=fakes.API_VERSION,
            allow_api_version_downgrade=False,
            session=instance.session,
            region_name=instance._region_name,
            endpoint_override='endpoint')
        instance.get_endpoint_for_service_type.assert_called_once_with(
            'baremetal', region_name=instance._region_name,
            interface=instance.interface)

    @mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=True)
    @mock.patch.object(client, 'Client', autospec=True)
    def test_make_client_latest(self, mock_client: mock.Mock) -> None:
        instance = fakes.FakeClientManager()
        instance.get_endpoint_for_service_type = mock.Mock(
            return_value='endpoint')
        instance._api_version = {'baremetal': plugin.LATEST_VERSION}
        plugin.make_client(instance)
        mock_client.assert_called_once_with(
            # NOTE(dtantsur): "latest" is changed to an actual version before
            # make_client is called.
            os_ironic_api_version=plugin.LATEST_VERSION,
            allow_api_version_downgrade=True,
            session=instance.session,
            region_name=instance._region_name,
            endpoint_override='endpoint')
        instance.get_endpoint_for_service_type.assert_called_once_with(
            'baremetal', region_name=instance._region_name,
            interface=instance.interface)

    @mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=False)
    @mock.patch.object(client, 'Client', autospec=True)
    def test_make_client_v1(self, mock_client: mock.Mock) -> None:
        instance = fakes.FakeClientManager()
        instance.get_endpoint_for_service_type = mock.Mock(
            return_value='endpoint')
        instance._api_version = {'baremetal': '1'}
        plugin.make_client(instance)
        mock_client.assert_called_once_with(
            os_ironic_api_version=plugin.LATEST_VERSION,
            allow_api_version_downgrade=True,
            session=instance.session,
            region_name=instance._region_name,
            endpoint_override='endpoint')
        instance.get_endpoint_for_service_type.assert_called_once_with(
            'baremetal', region_name=instance._region_name,
            interface=instance.interface)


@mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=True)
@mock.patch.object(argparse.ArgumentParser, 'add_argument', autospec=True)
class BuildOptionParserTest(testtools.TestCase):

    @mock.patch.object(plugin.utils, 'env', lambda x: None)
    def test_build_option_parser(self, mock_add_argument: mock.Mock) -> None:
        parser = argparse.ArgumentParser()
        mock_add_argument.reset_mock()
        plugin.build_option_parser(parser)
        version_list = ['1'] + ['1.%d' % i for i in range(
            1, plugin.LAST_KNOWN_API_VERSION + 1)] + ['latest']
        mock_add_argument.assert_called_once_with(
            mock.ANY, '--os-baremetal-api-version',
            action=plugin.ReplaceLatestVersion, choices=version_list,
            default=plugin.LATEST_VERSION, help=mock.ANY,
            metavar='<baremetal-api-version>')
        self.assertTrue(plugin.OS_BAREMETAL_API_LATEST)

    @mock.patch.object(plugin.utils, 'env', lambda x: "latest")
    def test_build_option_parser_env_latest(
            self, mock_add_argument: mock.Mock) -> None:
        parser = argparse.ArgumentParser()
        mock_add_argument.reset_mock()
        plugin.build_option_parser(parser)
        version_list = ['1'] + ['1.%d' % i for i in range(
            1, plugin.LAST_KNOWN_API_VERSION + 1)] + ['latest']
        mock_add_argument.assert_called_once_with(
            mock.ANY, '--os-baremetal-api-version',
            action=plugin.ReplaceLatestVersion, choices=version_list,
            default=plugin.LATEST_VERSION, help=mock.ANY,
            metavar='<baremetal-api-version>')
        self.assertTrue(plugin.OS_BAREMETAL_API_LATEST)

    @mock.patch.object(plugin.utils, 'env', lambda x: "1.1")
    def test_build_option_parser_env(
            self, mock_add_argument: mock.Mock) -> None:
        parser = argparse.ArgumentParser()
        mock_add_argument.reset_mock()
        plugin.build_option_parser(parser)
        version_list = ['1'] + ['1.%d' % i for i in range(
            1, plugin.LAST_KNOWN_API_VERSION + 1)] + ['latest']
        mock_add_argument.assert_called_once_with(
            mock.ANY, '--os-baremetal-api-version',
            action=plugin.ReplaceLatestVersion, choices=version_list,
            default='1.1', help=mock.ANY,
            metavar='<baremetal-api-version>')
        self.assertFalse(plugin.OS_BAREMETAL_API_LATEST)


@mock.patch.object(plugin.utils, 'env', lambda x: None)
class ReplaceLatestVersionTest(testtools.TestCase):

    @mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=False)
    def test___call___latest(self) -> None:
        parser = argparse.ArgumentParser()
        plugin.build_option_parser(parser)
        namespace = argparse.Namespace()
        parser.parse_known_args(['--os-baremetal-api-version', 'latest'],
                                namespace)
        self.assertEqual(plugin.LATEST_VERSION,
                         namespace.os_baremetal_api_version)
        self.assertTrue(plugin.OS_BAREMETAL_API_LATEST)

    @mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=True)
    def test___call___specific_version(self) -> None:
        parser = argparse.ArgumentParser()
        plugin.build_option_parser(parser)
        namespace = argparse.Namespace()
        parser.parse_known_args(['--os-baremetal-api-version', '1.4'],
                                namespace)
        self.assertEqual('1.4', namespace.os_baremetal_api_version)
        self.assertFalse(plugin.OS_BAREMETAL_API_LATEST)

    @mock.patch.object(plugin, 'OS_BAREMETAL_API_LATEST', new=True)
    def test___call___default(self) -> None:
        parser = argparse.ArgumentParser()
        plugin.build_option_parser(parser)
        namespace = argparse.Namespace()
        parser.parse_known_args([], namespace)
        self.assertEqual(plugin.LATEST_VERSION,
                         namespace.os_baremetal_api_version)
        self.assertTrue(plugin.OS_BAREMETAL_API_LATEST)



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import functools
from http import client as http_client
import json
import logging
import re
import textwrap
import threading
import time
from typing import Any, Callable, TypeVar
from urllib import parse as urlparse

from keystoneauth1 import adapter
from keystoneauth1 import exceptions as kexc
from keystoneauth1 import session as ks_session
import requests

from ironicclient.common import filecache
from ironicclient.common.i18n import _
from ironicclient import exc


# NOTE(deva): Record the latest version that this client was tested with.
#             We still have a lot of work to do in the client to implement
#             microversion support in the client properly! See
#             http://specs.openstack.org/openstack/ironic-specs/specs/kilo/api-microversions.html # noqa
#             for full details.
DEFAULT_VER: str = '1.9'
LAST_KNOWN_API_VERSION: int = 112
LATEST_VERSION: str = '1.{}'.format(LAST_KNOWN_API_VERSION)

LOG: logging.Logger = logging.getLogger(__name__)
USER_AGENT: str = 'python-ironicclient'
CHUNKSIZE: int = 1024 * 64  # 64kB

_MAJOR_VERSION: int = 1
API_VERSION: str = '/v%d' % _MAJOR_VERSION
API_VERSION_SELECTED_STATES: tuple[str, ...] = (
    'user', 'negotiated', 'cached', 'default')

DEFAULT_MAX_RETRIES: int = 5
DEFAULT_RETRY_INTERVAL: int = 2
SENSITIVE_HEADERS: tuple[str, ...] = ('X-Auth-Token',)

SUPPORTED_ENDPOINT_SCHEME: tuple[str, ...] = ('http', 'https')

_API_VERSION_RE: re.Pattern[str] = re.compile(r'/+(v%d)?/*$' % _MAJOR_VERSION)


@functools.total_ordering
class _Version:
    _version_re: re.Pattern[str] = re.compile(
        r'^(\d) \. (\d+)$', re.VERBOSE | re.ASCII)

    def __init__(self, version: str) -> None:
        match = self._version_re.match(version)
        if not match:
            raise ValueError('invalid version number %s' % version)
        major, minor = match.group(1, 2)
        self.version: tuple[int, int] = (int(major), int(minor))

    def __str__(self) -> str:
        return '.'.join(str(v) for v in self.version)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, _Version):
            return NotImplemented
        return self.version == other.version

    def __lt__(self, other: _Version) -> bool:
        return self.version < other.version


def _trim_endpoint_api_version(url: str) -> str:
    """Trim API version and trailing slash from endpoint."""
    return re.sub(_API_VERSION_RE, '', url)


def _extract_error_json(body: bytes | str) -> dict[str, Any]:
    """Return  error_message from the HTTP response body."""
    try:
        body_json: dict[str, Any] = json.loads(body)
    except ValueError:
        return {}

    if 'error_message' not in body_json:
        return {}

    try:
        error_json = json.loads(body_json['error_message'])
    except ValueError:
        return body_json

    err_msg = (error_json.get('faultstring') or error_json.get('description'))
    if err_msg:
        body_json['error_message'] = err_msg

    return body_json


def get_server(
    url: str | None,
) -> tuple[str | None, str | None]:
    """Extract and return the server & port."""
    if url is None:
        return None, None
    parts = urlparse.urlparse(url)
    return parts.hostname, str(parts.port)


class VersionNegotiationMixin(object):
    os_ironic_api_version: str | list[str]
    api_version_select_state: str

    def negotiate_version(
        self,
        conn: ks_session.Session,
        resp: requests.Response | None,
    ) -> str:
        """Negotiate the server version

        Assumption: Called after receiving a 406 error when doing a request.

        :param conn: A connection object
        :param resp: The response object from http request
        """
        def _query_server(conn: ks_session.Session) -> requests.Response:
            if (self.os_ironic_api_version
                    and not isinstance(self.os_ironic_api_version, list)
                    and self.os_ironic_api_version != 'latest'):
                base_version = ("/v%s" %
                                str(self.os_ironic_api_version).split('.')[0])
            else:
                base_version = API_VERSION
            # Raise exception on client or server error.
            resp = self._make_simple_request(conn, 'GET', base_version)
            if not resp.ok:
                raise exc.from_response(resp, method='GET', url=base_version)
            return resp

        version_overridden = False

        if (resp and hasattr(resp, 'request')
                and hasattr(resp.request, 'headers')):
            orig_hdr = resp.request.headers
            # Get the version of the client's last request and fallback
            # to the default for things like unit tests to not cause
            # migraines.
            req_api_ver = orig_hdr.get('X-OpenStack-Ironic-API-Version',
                                       self.os_ironic_api_version)
        else:
            req_api_ver = self.os_ironic_api_version
        if (resp and req_api_ver != self.os_ironic_api_version
                and self.api_version_select_state == 'negotiated'):
            # If we have a non-standard api version on the request,
            # but we think we've negotiated, then the call was overridden.
            # We should report the error with the called version
            requested_version = req_api_ver
            # And then we shouldn't save the newly negotiated
            # version of this negotiation because we have been
            # overridden a request.
            version_overridden = True
        else:
            requested_version = self.os_ironic_api_version

        if not resp:
            resp = _query_server(conn)
        if self.api_version_select_state not in API_VERSION_SELECTED_STATES:
            raise RuntimeError(
                _('Error: self.api_version_select_state should be one of the '
                  'values in: "%(valid)s" but had the value: "%(value)s"') %
                {'valid': ', '.join(API_VERSION_SELECTED_STATES),
                 'value': self.api_version_select_state})
        min_ver, max_ver = self._parse_version_headers(resp)
        # NOTE: servers before commit 32fb6e99 did not return version headers
        # on error, so we need to perform a GET to determine
        # the supported version range
        if not max_ver:
            LOG.debug('No version header in response, requesting from server')
            resp = _query_server(conn)
            min_ver, max_ver = self._parse_version_headers(resp)
        # Reset the maximum version that we permit
        if min_ver is None or max_ver is None:
            raise RuntimeError(
                _('Could not determine API version range from server; '
                  'version headers missing from response.'))

        if _Version(max_ver) > _Version(LATEST_VERSION):
            LOG.debug("Remote API version %(max_ver)s is greater than the "
                      "version supported by ironicclient. Maximum available "
                      "version is %(client_ver)s",
                      {'max_ver': max_ver,
                       'client_ver': LATEST_VERSION})
            max_ver = LATEST_VERSION

        # If the user requested an explicit version or we have negotiated a
        # version and still failing then error now.  The server could
        # support the version requested but the requested operation may not
        # be supported by the requested version.
        # TODO(TheJulia): We should break this method into several parts,
        # such as a sanity check/error method.
        if ((self.api_version_select_state == 'user'
             and not self._must_negotiate_version())
                or (self.api_version_select_state == 'negotiated'
                    and version_overridden)):
            raise exc.UnsupportedVersion(textwrap.fill(
                _("Requested API version %(req)s is not supported by the "
                  "server, client, or the requested operation is not "
                  "supported by the requested version. "
                  "Supported version range is %(min)s to "
                  "%(max)s")
                % {'req': requested_version,
                   'min': min_ver, 'max': max_ver}))
        if (self.api_version_select_state == 'negotiated'):
            raise exc.UnsupportedVersion(textwrap.fill(
                _("No API version was specified or the requested operation "
                  "was not supported by the client's negotiated API version "
                  "%(req)s.  Supported version range is: %(min)s to %(max)s")
                % {'req': requested_version,
                   'min': min_ver, 'max': max_ver}))

        if isinstance(requested_version, str):
            if requested_version == 'latest':
                negotiated_ver = max_ver
            else:
                negotiated_ver = str(
                    min(_Version(requested_version), _Version(max_ver))
                )

        elif isinstance(requested_version, list):
            if 'latest' in requested_version:
                raise ValueError(textwrap.fill(
                    _("The 'latest' API version can not be requested "
                      "in a list of versions. Please explicitly request "
                      "'latest' or request only versions between "
                      "%(min)s to %(max)s")
                    % {'min': min_ver, 'max': max_ver}))

            versions = []
            for version in requested_version:
                if _Version(min_ver) <= _Version(version) <= _Version(max_ver):
                    versions.append(_Version(version))
            if versions:
                negotiated_ver = str(max(versions))
            else:
                raise exc.UnsupportedVersion(textwrap.fill(
                    _("Requested API version specified and the requested "
                      "operation was not supported by the client's "
                      "requested API version %(req)s.  Supported "
                      "version range is: %(min)s to %(max)s")
                    % {'req': requested_version,
                       'min': min_ver, 'max': max_ver}))

        else:
            raise ValueError(textwrap.fill(
                _("Requested API version %(req)s type is unsupported. "
                  "Valid types are Strings such as '1.1', 'latest' "
                  "or a list of string values representing API versions.")
                % {'req': requested_version}))

        if _Version(negotiated_ver) < _Version(min_ver):
            negotiated_ver = min_ver

        # server handles microversions, but doesn't support
        # the requested version, so try a negotiated version
        self.api_version_select_state = 'negotiated'
        self.os_ironic_api_version = negotiated_ver
        LOG.debug('Negotiated API version is %s', negotiated_ver)

        # Cache the negotiated version for this server
        endpoint_override = getattr(self, 'endpoint_override', None)
        host, port = get_server(endpoint_override)
        if host is not None and port is not None:
            filecache.save_data(host=host, port=port, data=negotiated_ver)

        return negotiated_ver

    def _generic_parse_version_headers(
        self,
        accessor_func: Callable[..., str | None],
    ) -> tuple[str | None, str | None]:
        min_ver = accessor_func('X-OpenStack-Ironic-API-Minimum-Version', None)
        max_ver = accessor_func('X-OpenStack-Ironic-API-Maximum-Version', None)
        return min_ver, max_ver

    def _parse_version_headers(
        self,
        resp: requests.Response,
    ) -> tuple[str | None, str | None]:
        # NOTE(jlvillal): Declared for unit testing purposes
        raise NotImplementedError()

    def _make_simple_request(
        self,
        conn: ks_session.Session,
        method: str,
        url: str,
    ) -> requests.Response:
        # NOTE(jlvillal): Declared for unit testing purposes
        raise NotImplementedError()

    def _must_negotiate_version(self) -> bool:
        return (self.api_version_select_state == 'user'
                and (self.os_ironic_api_version == 'latest'
                     or isinstance(self.os_ironic_api_version, list)))


_RETRY_EXCEPTIONS: tuple[type[Exception], ...] = (
    exc.Conflict, exc.ServiceUnavailable,
    exc.ConnectionRefused, kexc.RetriableConnectionFailure)

_RT = TypeVar('_RT')


def with_retries(
    func: Callable[..., _RT],
) -> Callable[..., _RT]:
    """Wrapper for _http_request adding support for retries."""
    @functools.wraps(func)
    def wrapper(
        self: SessionClient,
        url: str,
        method: str,
        **kwargs: Any,
    ) -> _RT:
        if self.conflict_max_retries is None:
            self.conflict_max_retries = DEFAULT_MAX_RETRIES
        if self.conflict_retry_interval is None:
            self.conflict_retry_interval = DEFAULT_RETRY_INTERVAL

        num_attempts = self.conflict_max_retries + 1
        for attempt in range(1, num_attempts + 1):
            try:
                return func(self, url, method, **kwargs)
            except _RETRY_EXCEPTIONS as error:
                msg = ("Error contacting Ironic server: %(error)s. "
                       "Attempt %(attempt)d of %(total)d" %
                       {'attempt': attempt,
                        'total': num_attempts,
                        'error': error})
                if attempt == num_attempts:
                    LOG.error(msg)
                    raise
                else:
                    LOG.debug(msg)
                    time.sleep(self.conflict_retry_interval)
        raise AssertionError('unreachable')

    return wrapper


class SessionClient(VersionNegotiationMixin, adapter.LegacyJsonAdapter):
    """HTTP client based on Keystone client session."""

    def __init__(
        self,
        os_ironic_api_version: str | list[str],
        api_version_select_state: str,
        max_retries: int | None,
        retry_interval: int | None,
        **kwargs: Any,
    ) -> None:
        self.os_ironic_api_version = os_ironic_api_version
        self.api_version_select_state = api_version_select_state
        self.conflict_max_retries = max_retries
        self.conflict_retry_interval = retry_interval
        if isinstance(kwargs.get('endpoint_override'), str):
            kwargs['endpoint_override'] = _trim_endpoint_api_version(
                kwargs['endpoint_override'])

        super(SessionClient, self).__init__(**kwargs)

        endpoint_filter = self._get_endpoint_filter()
        # type of endpoint_filter is incompatible with ksa stubs
        endpoint = self.get_endpoint(
            **endpoint_filter  # type: ignore[arg-type]
        )
        if endpoint is None:
            raise exc.EndpointNotFound(
                _('The Bare Metal API endpoint cannot be detected and was '
                  'not provided explicitly'))
        self.endpoint_trimmed = _trim_endpoint_api_version(endpoint)
        self._first_negotiation_lock = threading.Lock()

    def _parse_version_headers(
        self,
        resp: requests.Response,
    ) -> tuple[str | None, str | None]:
        return self._generic_parse_version_headers(resp.headers.get)

    def _get_endpoint_filter(self) -> dict[str, str | list[str] | None]:
        return {
            'interface': self.interface,
            'service_type': self.service_type,
            'region_name': self.region_name
        }

    def _make_simple_request(
        self,
        conn: ks_session.Session,
        method: str,
        url: str,
    ) -> requests.Response:
        # NOTE: conn is self.session for this class
        return conn.request(url, method, raise_exc=False,
                            user_agent=USER_AGENT,
                            endpoint_filter=self._get_endpoint_filter(),
                            endpoint_override=self.endpoint_override)

    @with_retries
    def _http_request(
        self,
        url: str,
        method: str,
        **kwargs: Any,
    ) -> requests.Response:

        # NOTE(TheJulia): self.os_ironic_api_version is reset in
        # the self.negotiate_version() call if negotiation occurs.
        if self.os_ironic_api_version and self._must_negotiate_version():
            with self._first_negotiation_lock:
                if self._must_negotiate_version():
                    self.negotiate_version(self.session, None)

        kwargs.setdefault('user_agent', USER_AGENT)
        kwargs.setdefault('auth', self.auth)
        if isinstance(self.endpoint_override, str):
            kwargs.setdefault('endpoint_override', self.endpoint_override)

        if getattr(self, 'os_ironic_api_version', None):
            kwargs['headers'].setdefault('X-OpenStack-Ironic-API-Version',
                                         self.os_ironic_api_version)

        for k, v in self.additional_headers.items():
            kwargs['headers'].setdefault(k, v)

        if self.global_request_id is not None:
            kwargs['headers'].setdefault(
                "X-OpenStack-Request-ID", self.global_request_id)

        endpoint_filter = kwargs.setdefault('endpoint_filter', {})
        endpoint_filter.setdefault('interface', self.interface)
        endpoint_filter.setdefault('service_type', self.service_type)
        endpoint_filter.setdefault('region_name', self.region_name)

        resp = self.session.request(url, method,
                                    raise_exc=False, **kwargs)
        if resp.status_code == http_client.NOT_ACCEPTABLE:
            negotiated_ver = self.negotiate_version(self.session, resp)
            kwargs['headers']['X-OpenStack-Ironic-API-Version'] = (
                negotiated_ver)
            return self._http_request(url, method, **kwargs)
        if resp.status_code >= http_client.BAD_REQUEST:
            error_json = _extract_error_json(resp.content)
            # exc.from_response expects str|None; error_json is dict[str, Any]
            raise exc.from_response(  # type: ignore[call-arg]
                resp,
                error_json.get('error_message'),  # type: ignore[arg-type]
                error_json.get('debuginfo'),  # type: ignore[arg-type]
                method,
                url,
            )
        elif resp.status_code in (http_client.MOVED_PERMANENTLY,
                                  http_client.FOUND, http_client.USE_PROXY):
            # Redirected. Reissue the request to the new location.
            location = resp.headers.get('location')
            resp = self._http_request(location, method, **kwargs)
        elif resp.status_code == http_client.MULTIPLE_CHOICES:
            raise exc.from_response(resp, method=method, url=url)
        return resp

    def json_request(
        self,
        method: str,
        url: str,
        **kwargs: Any,
    ) -> tuple[requests.Response, dict[str, Any] | list[Any] | bytes | None]:
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type', 'application/json')
        kwargs['headers'].setdefault('Accept', 'application/json')

        if 'body' in kwargs:
            kwargs['json'] = kwargs.pop('body')

        resp = self._http_request(url, method, **kwargs)
        body = resp.content
        content_type = resp.headers.get('content-type', None)
        status = resp.status_code
        if (status in (http_client.NO_CONTENT, http_client.RESET_CONTENT)
                or content_type is None):
            return resp, list()
        if 'application/json' in content_type:
            try:
                body = resp.json()
            except ValueError:
                LOG.error('Could not decode response body as JSON')
        else:
            body = None

        return resp, body

    def raw_request(
        self,
        method: str,
        url: str,
        **kwargs: dict[str, str],
    ) -> requests.Response:
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type',
                                     'application/octet-stream')
        return self._http_request(url, method, **kwargs)


def _construct_http_client(
    session: ks_session.Session,
    token: str | None = None,
    auth_ref: object = None,
    os_ironic_api_version: str | list[str] = DEFAULT_VER,
    api_version_select_state: str = 'default',
    max_retries: int = DEFAULT_MAX_RETRIES,
    retry_interval: int = DEFAULT_RETRY_INTERVAL,
    timeout: int = 600,
    ca_file: str | None = None,
    cert_file: str | None = None,
    key_file: str | None = None,
    insecure: bool | None = None,
    **kwargs: Any,
) -> SessionClient:

    kwargs.setdefault('service_type', 'baremetal')
    kwargs.setdefault('user_agent', 'python-ironicclient')
    kwargs.setdefault('interface', kwargs.pop('endpoint_type',
                                              'publicURL'))

    ignored = {'token': token,
               'auth_ref': auth_ref,
               'timeout': timeout != 600,
               'ca_file': ca_file,
               'cert_file': cert_file,
               'key_file': key_file,
               'insecure': insecure}

    dvars = [k for k, v in ignored.items() if v]

    if dvars:
        LOG.warning('The following arguments are ignored when using '
                    'the session to construct a client: %s',
                    ', '.join(dvars))

    return SessionClient(session=session,
                         os_ironic_api_version=os_ironic_api_version,
                         api_version_select_state=api_version_select_state,
                         max_retries=max_retries,
                         retry_interval=retry_interval,
                         **kwargs)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from __future__ import annotations

import copy
from unittest import mock

from osc_lib.tests import utils as osctestutils

from ironicclient import exc
from ironicclient.osc.v1 import baremetal_allocation
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalAllocation(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestBaremetalAllocation, self).setUp()

        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestCreateBaremetalAllocation(TestBaremetalAllocation):

    def setUp(self) -> None:
        super(TestCreateBaremetalAllocation, self).setUp()

        self.baremetal_mock.allocation.create.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True,
            ))

        self.baremetal_mock.allocation.wait.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_allocation.CreateBaremetalAllocation(self.app,
                                                                  None)

    def test_baremetal_allocation_create(self) -> None:
        arglist = [
            '--resource-class', baremetal_fakes.baremetal_resource_class,
        ]

        verifylist = [
            ('resource_class', baremetal_fakes.baremetal_resource_class),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)

    def test_baremetal_allocation_create_wait(self) -> None:
        arglist = [
            '--resource-class', baremetal_fakes.baremetal_resource_class,
            '--wait',
        ]

        verifylist = [
            ('resource_class', baremetal_fakes.baremetal_resource_class),
            ('wait_timeout', 0),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)
        self.baremetal_mock.allocation.wait.assert_called_once_with(
            baremetal_fakes.ALLOCATION['uuid'], timeout=0)

    def test_baremetal_allocation_create_wait_with_timeout(self) -> None:
        arglist = [
            '--resource-class', baremetal_fakes.baremetal_resource_class,
            '--wait', '3600',
        ]

        verifylist = [
            ('resource_class', baremetal_fakes.baremetal_resource_class),
            ('wait_timeout', 3600),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)
        self.baremetal_mock.allocation.wait.assert_called_once_with(
            baremetal_fakes.ALLOCATION['uuid'], timeout=3600)

    def test_baremetal_allocation_create_name_extras(self) -> None:
        arglist = [
            '--resource-class', baremetal_fakes.baremetal_resource_class,
            '--uuid', baremetal_fakes.baremetal_uuid,
            '--name', baremetal_fakes.baremetal_name,
            '--extra', 'key1=value1',
            '--extra', 'key2=value2'
        ]

        verifylist = [
            ('resource_class', baremetal_fakes.baremetal_resource_class),
            ('uuid', baremetal_fakes.baremetal_uuid),
            ('name', baremetal_fakes.baremetal_name),
            ('extra', ['key1=value1', 'key2=value2'])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
            'uuid': baremetal_fakes.baremetal_uuid,
            'name': baremetal_fakes.baremetal_name,
            'extra': {'key1': 'value1', 'key2': 'value2'}
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)

    def test_baremetal_allocation_create_nodes_and_traits(self) -> None:
        arglist = [
            '--resource-class', baremetal_fakes.baremetal_resource_class,
            '--candidate-node', 'node1',
            '--trait', 'CUSTOM_1',
            '--candidate-node', 'node2',
            '--trait', 'CUSTOM_2',
        ]

        verifylist = [
            ('resource_class', baremetal_fakes.baremetal_resource_class),
            ('candidate_nodes', ['node1', 'node2']),
            ('traits', ['CUSTOM_1', 'CUSTOM_2']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
            'candidate_nodes': ['node1', 'node2'],
            'traits': ['CUSTOM_1', 'CUSTOM_2'],
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)

    def test_baremetal_allocation_create_owner(self) -> None:
        arglist = [
            '--resource-class', baremetal_fakes.baremetal_resource_class,
            '--owner', baremetal_fakes.baremetal_owner,
        ]

        verifylist = [
            ('resource_class', baremetal_fakes.baremetal_resource_class),
            ('owner', baremetal_fakes.baremetal_owner),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
            'owner': baremetal_fakes.baremetal_owner,
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)

    def test_baremetal_allocation_create_no_options(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(exc.ClientException,
                          self.cmd.take_action,
                          parsed_args)

    def test_baremetal_allocation_backfill(self) -> None:
        arglist = [
            '--node', baremetal_fakes.baremetal_uuid,
        ]

        verifylist = [
            ('node', baremetal_fakes.baremetal_uuid),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        args = {
            'node': baremetal_fakes.baremetal_uuid,
        }

        self.baremetal_mock.allocation.create.assert_called_once_with(**args)


class TestShowBaremetalAllocation(TestBaremetalAllocation):

    def setUp(self) -> None:
        super(TestShowBaremetalAllocation, self).setUp()

        self.baremetal_mock.allocation.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True))

        self.cmd = baremetal_allocation.ShowBaremetalAllocation(self.app, None)

    def test_baremetal_allocation_show(self) -> None:
        arglist = [baremetal_fakes.baremetal_uuid]
        verifylist = [('allocation', baremetal_fakes.baremetal_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.baremetal_mock.allocation.get.assert_called_once_with(
            baremetal_fakes.baremetal_uuid, fields=None)

        collist = ('name', 'node_uuid', 'resource_class', 'state', 'uuid')
        self.assertEqual(collist, columns)

        datalist = (
            baremetal_fakes.baremetal_name,
            baremetal_fakes.baremetal_uuid,
            baremetal_fakes.baremetal_resource_class,
            baremetal_fakes.baremetal_allocation_state,
            baremetal_fakes.baremetal_uuid,
        )

        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_show_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalAllocationList(TestBaremetalAllocation):
    def setUp(self) -> None:
        super(TestBaremetalAllocationList, self).setUp()

        self.baremetal_mock.allocation.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True)
        ]
        self.cmd = baremetal_allocation.ListBaremetalAllocation(self.app, None)

    def test_baremetal_allocation_list(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None}
        self.baremetal_mock.allocation.list.assert_called_with(**kwargs)

        collist = (
            "uuid",
            "name",
            "resource_class",
            "state",
            "node_uuid")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_name,
                     baremetal_fakes.baremetal_resource_class,
                     baremetal_fakes.baremetal_allocation_state,
                     baremetal_fakes.baremetal_uuid),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_list_node(self) -> None:
        arglist = ['--node', baremetal_fakes.baremetal_uuid]
        verifylist = [('node', baremetal_fakes.baremetal_uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'node': baremetal_fakes.baremetal_uuid,
            'marker': None,
            'limit': None}
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

        collist = (
            "uuid",
            "name",
            "resource_class",
            "state",
            "node_uuid")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_name,
                     baremetal_fakes.baremetal_resource_class,
                     baremetal_fakes.baremetal_allocation_state,
                     baremetal_fakes.baremetal_uuid),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_list_resource_class(self) -> None:
        arglist = ['--resource-class',
                   baremetal_fakes.baremetal_resource_class]
        verifylist = [('resource_class',
                       baremetal_fakes.baremetal_resource_class)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'resource_class': baremetal_fakes.baremetal_resource_class,
            'marker': None,
            'limit': None}
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

        collist = (
            "uuid",
            "name",
            "resource_class",
            "state",
            "node_uuid")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_name,
                     baremetal_fakes.baremetal_resource_class,
                     baremetal_fakes.baremetal_allocation_state,
                     baremetal_fakes.baremetal_uuid),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_list_owner(self) -> None:
        arglist = ['--owner',
                   baremetal_fakes.baremetal_owner]
        verifylist = [('owner',
                       baremetal_fakes.baremetal_owner)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'owner': baremetal_fakes.baremetal_owner,
            'marker': None,
            'limit': None}
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

        collist = (
            "uuid",
            "name",
            "resource_class",
            "state",
            "node_uuid")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_name,
                     baremetal_fakes.baremetal_resource_class,
                     baremetal_fakes.baremetal_allocation_state,
                     baremetal_fakes.baremetal_uuid),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_list_state(self) -> None:
        arglist = ['--state', baremetal_fakes.baremetal_allocation_state]
        verifylist = [('state', baremetal_fakes.baremetal_allocation_state)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'state': baremetal_fakes.baremetal_allocation_state,
            'marker': None,
            'limit': None}
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

        collist = (
            "uuid",
            "name",
            "resource_class",
            "state",
            "node_uuid")
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_name,
                     baremetal_fakes.baremetal_resource_class,
                     baremetal_fakes.baremetal_allocation_state,
                     baremetal_fakes.baremetal_uuid),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_list_long(self) -> None:
        arglist = ['--long']
        verifylist = [('long', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
        }
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

        collist = ('uuid',
                   'name',
                   'state',
                   'owner',
                   'node_uuid',
                   'last_error',
                   'resource_class',
                   'traits',
                   'candidate_nodes',
                   'extra',
                   'created_at',
                   'updated_at')
        self.assertEqual(collist, columns)

        datalist = ((baremetal_fakes.baremetal_uuid,
                     baremetal_fakes.baremetal_name,
                     baremetal_fakes.baremetal_allocation_state,
                     '',
                     baremetal_fakes.baremetal_uuid,
                     '',
                     baremetal_fakes.baremetal_resource_class,
                     '',
                     '',
                     '',
                     '',
                     ''),)
        self.assertEqual(datalist, tuple(data))

    def test_baremetal_allocation_list_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'node_uuid']
        verifylist = [('fields', [['uuid', 'node_uuid']])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
            'fields': ('uuid', 'node_uuid')
        }
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

    def test_baremetal_allocation_list_fields_multiple(self) -> None:
        arglist = ['--fields', 'uuid', 'node_uuid', '--fields', 'extra']
        verifylist = [('fields', [['uuid', 'node_uuid'], ['extra']])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        kwargs = {
            'marker': None,
            'limit': None,
            'fields': ('uuid', 'node_uuid', 'extra')
        }
        self.baremetal_mock.allocation.list.assert_called_once_with(**kwargs)

    def test_baremetal_allocation_list_invalid_fields(self) -> None:
        arglist = ['--fields', 'uuid', 'invalid']
        verifylist = [('fields', [['uuid', 'invalid']])]
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalAllocationDelete(TestBaremetalAllocation):

    def setUp(self) -> None:
        super(TestBaremetalAllocationDelete, self).setUp()

        self.baremetal_mock.allocation.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True))

        self.cmd = baremetal_allocation.DeleteBaremetalAllocation(self.app,
                                                                  None)

    def test_baremetal_allocation_delete(self) -> None:
        arglist = [baremetal_fakes.baremetal_uuid]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.baremetal_mock.allocation.delete.assert_called_once_with(
            baremetal_fakes.baremetal_uuid)

    def test_baremetal_allocation_delete_multiple(self) -> None:
        arglist = [baremetal_fakes.baremetal_uuid,
                   baremetal_fakes.baremetal_name]
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.baremetal_mock.allocation.delete.assert_has_calls(
            [mock.call(x) for x in arglist]
        )
        self.assertEqual(2, self.baremetal_mock.allocation.delete.call_count)

    def test_baremetal_allocation_delete_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalAllocationSet(TestBaremetalAllocation):
    def setUp(self) -> None:
        super(TestBaremetalAllocationSet, self).setUp()

        self.baremetal_mock.allocation.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True))

        self.cmd = baremetal_allocation.SetBaremetalAllocation(
            self.app, None)

    def test_baremetal_allocation_set_name(self) -> None:
        new_name = 'foo'
        arglist = [
            baremetal_fakes.baremetal_uuid,
            '--name', new_name]
        verifylist = [
            ('allocation', baremetal_fakes.baremetal_uuid),
            ('name', new_name)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.allocation.update.assert_called_once_with(
            baremetal_fakes.baremetal_uuid,
            [{'path': '/name', 'value': new_name, 'op': 'add'}])

    def test_baremetal_allocation_set_extra(self) -> None:
        extra_value = 'foo=bar'
        arglist = [
            baremetal_fakes.baremetal_uuid,
            '--extra', extra_value]
        verifylist = [
            ('allocation', baremetal_fakes.baremetal_uuid),
            ('extra', [extra_value])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.allocation.update.assert_called_once_with(
            baremetal_fakes.baremetal_uuid,
            [{'path': '/extra/foo', 'value': 'bar', 'op': 'add'}])

    def test_baremetal_allocation_set_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalAllocationUnset(TestBaremetalAllocation):
    def setUp(self) -> None:
        super(TestBaremetalAllocationUnset, self).setUp()

        self.baremetal_mock.allocation.update.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.ALLOCATION),
                loaded=True))

        self.cmd = baremetal_allocation.UnsetBaremetalAllocation(
            self.app, None)

    def test_baremetal_allocation_unset_name(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_uuid, '--name']
        verifylist = [('allocation',
                       baremetal_fakes.baremetal_uuid),
                      ('name', True)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.allocation.update.assert_called_once_with(
            baremetal_fakes.baremetal_uuid,
            [{'path': '/name', 'op': 'remove'}])

    def test_baremetal_allocation_unset_extra(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_uuid, '--extra', 'key1']
        verifylist = [('allocation',
                       baremetal_fakes.baremetal_uuid),
                      ('extra', ['key1'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.allocation.update.assert_called_once_with(
            baremetal_fakes.baremetal_uuid,
            [{'path': '/extra/key1', 'op': 'remove'}])

    def test_baremetal_allocation_unset_multiple_extras(self) -> None:
        arglist = [
            baremetal_fakes.baremetal_uuid,
            '--extra', 'key1', '--extra', 'key2']
        verifylist = [('allocation',
                       baremetal_fakes.baremetal_uuid),
                      ('extra', ['key1', 'key2'])]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)
        self.baremetal_mock.allocation.update.assert_called_once_with(
            baremetal_fakes.baremetal_uuid,
            [{'path': '/extra/key1', 'op': 'remove'},
             {'path': '/extra/key2', 'op': 'remove'}])

    def test_baremetal_allocation_unset_no_options(self) -> None:
        arglist = []
        verifylist = []
        self.assertRaises(osctestutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_baremetal_allocation_unset_no_property(self) -> None:
        uuid = baremetal_fakes.baremetal_uuid
        arglist = [uuid]
        verifylist = [('allocation', uuid)]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)
        self.assertFalse(self.baremetal_mock.allocation.update.called)



# Copyright 2013 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

import builtins
import json
import os
import subprocess
import sys
import tempfile
from typing import Any
from unittest import mock


from ironicclient.common import utils
from ironicclient import exc
from ironicclient.tests.unit import utils as test_utils


class UtilsTest(test_utils.BaseTestCase):

    def test_key_value_pairs_to_dict(self) -> None:
        kv_list = ['str=foo', 'int=1', 'bool=true',
                   'list=[1, 2, 3]', 'dict={"foo": "bar"}']

        d = utils.key_value_pairs_to_dict(kv_list)
        self.assertEqual(
            {'str': 'foo', 'int': 1, 'bool': True,
             'list': [1, 2, 3], 'dict': {'foo': 'bar'}},
            d)

    def test_key_value_pairs_to_dict_nothing(self) -> None:
        self.assertEqual({}, utils.key_value_pairs_to_dict(None))
        self.assertEqual({}, utils.key_value_pairs_to_dict([]))

    def test_args_array_to_dict(self) -> None:
        my_args = {
            'matching_metadata': ['str=foo', 'int=1', 'bool=true',
                                  'list=[1, 2, 3]', 'dict={"foo": "bar"}'],
            'other': 'value'
        }
        cleaned_dict = utils.args_array_to_dict(my_args,
                                                "matching_metadata")
        self.assertEqual({
            'matching_metadata': {'str': 'foo', 'int': 1, 'bool': True,
                                  'list': [1, 2, 3], 'dict': {'foo': 'bar'}},
            'other': 'value'
        }, cleaned_dict)

    def test_args_array_to_patch(self) -> None:
        my_args = {
            'attributes': ['str=foo', 'int=1', 'bool=true',
                           'list=[1, 2, 3]', 'dict={"foo": "bar"}'],
            'op': 'add',
        }
        patch = utils.args_array_to_patch(my_args['op'],
                                          my_args['attributes'])
        self.assertEqual([{'op': 'add', 'value': 'foo', 'path': '/str'},
                          {'op': 'add', 'value': 1, 'path': '/int'},
                          {'op': 'add', 'value': True, 'path': '/bool'},
                          {'op': 'add', 'value': [1, 2, 3], 'path': '/list'},
                          {'op': 'add', 'value': {"foo": "bar"},
                           'path': '/dict'}], patch)

    def test_args_array_to_patch_format_error(self) -> None:
        my_args = {
            'attributes': ['foobar'],
            'op': 'add',
        }
        self.assertRaises(exc.CommandError, utils.args_array_to_patch,
                          my_args['op'], my_args['attributes'])

    def test_args_array_to_patch_remove(self) -> None:
        my_args = {
            'attributes': ['/foo', 'extra/bar'],
            'op': 'remove',
        }
        patch = utils.args_array_to_patch(my_args['op'],
                                          my_args['attributes'])
        self.assertEqual([{'op': 'remove', 'path': '/foo'},
                          {'op': 'remove', 'path': '/extra/bar'}], patch)

    def test_split_and_deserialize(self) -> None:
        ret = utils.split_and_deserialize('str=foo')
        self.assertEqual(('str', 'foo'), ret)

        ret = utils.split_and_deserialize('int=1')
        self.assertEqual(('int', 1), ret)

        ret = utils.split_and_deserialize('bool=false')
        self.assertEqual(('bool', False), ret)

        ret = utils.split_and_deserialize('list=[1, "foo", 2]')
        self.assertEqual(('list', [1, "foo", 2]), ret)

        ret = utils.split_and_deserialize('dict={"foo": 1}')
        self.assertEqual(('dict', {"foo": 1}), ret)

        ret = utils.split_and_deserialize('str_int="1"')
        self.assertEqual(('str_int', "1"), ret)

    def test_split_and_deserialize_fail(self) -> None:
        self.assertRaises(exc.CommandError,
                          utils.split_and_deserialize, 'foo:bar')

    def test_bool_arg_value(self) -> None:
        self.assertTrue(utils.bool_argument_value('arg', 'y', strict=True))
        self.assertTrue(utils.bool_argument_value('arg', 'TrUe', strict=True))
        self.assertTrue(utils.bool_argument_value('arg', '1', strict=True))
        self.assertTrue(utils.bool_argument_value('arg', 1, strict=True))

        self.assertFalse(utils.bool_argument_value('arg', '0', strict=True))
        self.assertFalse(utils.bool_argument_value('arg', 'No', strict=True))

        self.assertRaises(exc.CommandError,
                          utils.bool_argument_value, 'arg', 'ee', strict=True)

        self.assertFalse(utils.bool_argument_value('arg', 'ee', strict=False))
        self.assertTrue(utils.bool_argument_value('arg', 'ee', strict=False,
                                                  default=True))
        # No check that default is a Boolean...
        self.assertEqual('foo', utils.bool_argument_value('arg', 'ee',
                         strict=False, default='foo'))

    def test_check_for_invalid_fields(self) -> None:
        utils.check_for_invalid_fields(['a', 'b'], ['a', 'b', 'c'])
        # 'd' is not a valid field
        self.assertRaises(exc.CommandError, utils.check_for_invalid_fields,
                          ['a', 'd'], ['a', 'b', 'c'])

    def test_convert_list_props_to_comma_separated_strings(
        self,
    ) -> None:
        data = {'prop1': 'val1',
                'prop2': ['item1', 'item2', 'item3']}
        result = utils.convert_list_props_to_comma_separated(data)
        self.assertEqual('val1', result['prop1'])
        self.assertEqual('item1, item2, item3', result['prop2'])

    def test_convert_list_props_to_comma_separated_mix(
        self,
    ) -> None:
        data = {'prop1': 'val1',
                'prop2': [1, 2.5, 'item3']}
        result = utils.convert_list_props_to_comma_separated(data)
        self.assertEqual('val1', result['prop1'])
        self.assertEqual('1, 2.5, item3', result['prop2'])

    def test_convert_list_props_to_comma_separated_partial(
        self,
    ) -> None:
        data = {'prop1': [1, 2, 3],
                'prop2': [1, 2.5, 'item3']}
        result = utils.convert_list_props_to_comma_separated(
            data, props=['prop2'])
        self.assertEqual([1, 2, 3], result['prop1'])
        self.assertEqual('1, 2.5, item3', result['prop2'])


class CommonParamsForListTest(test_utils.BaseTestCase):
    def setUp(self) -> None:
        super(CommonParamsForListTest, self).setUp()
        self.args = mock.Mock(marker=None, limit=None, sort_key=None,
                              sort_dir=None, detail=False, fields=None,
                              spec=True)
        self.expected_params: dict[str, Any] = {'detail': False}

    def test_nothing_set(self) -> None:
        self.assertEqual(self.expected_params,
                         utils.common_params_for_list(self.args, [], []))

    def test_marker_and_limit(self) -> None:
        self.args.marker = 'foo'
        self.args.limit = 42
        self.expected_params.update({'marker': 'foo', 'limit': 42})
        self.assertEqual(self.expected_params,
                         utils.common_params_for_list(self.args, [], []))

    def test_invalid_limit(self) -> None:
        self.args.limit = -42
        self.assertRaises(exc.CommandError,
                          utils.common_params_for_list,
                          self.args, [], [])

    def test_sort_key_and_sort_dir(self) -> None:
        self.args.sort_key = 'field'
        self.args.sort_dir = 'desc'
        self.expected_params.update({'sort_key': 'field', 'sort_dir': 'desc'})
        self.assertEqual(self.expected_params,
                         utils.common_params_for_list(self.args,
                                                      ['field'],
                                                      []))

    def test_sort_key_allows_label(self) -> None:
        self.args.sort_key = 'Label'
        self.expected_params.update({'sort_key': 'field'})
        self.assertEqual(self.expected_params,
                         utils.common_params_for_list(self.args,
                                                      ['field', 'field2'],
                                                      ['Label', 'Label2']))

    def test_sort_key_invalid(self) -> None:
        self.args.sort_key = 'something'
        self.assertRaises(exc.CommandError,
                          utils.common_params_for_list,
                          self.args,
                          ['field', 'field2'],
                          [])

    def test_sort_dir_invalid(self) -> None:
        self.args.sort_dir = 'something'
        self.assertRaises(exc.CommandError,
                          utils.common_params_for_list,
                          self.args,
                          [],
                          [])

    def test_detail(self) -> None:
        self.args.detail = True
        self.expected_params['detail'] = True
        self.assertEqual(self.expected_params,
                         utils.common_params_for_list(self.args, [], []))

    def test_fields(self) -> None:
        self.args.fields = [['a', 'b', 'c']]
        self.expected_params.update({'fields': ['a', 'b', 'c']})
        self.assertEqual(self.expected_params,
                         utils.common_params_for_list(self.args, [], []))


class CommonFiltersTest(test_utils.BaseTestCase):
    def test_limit(self) -> None:
        result = utils.common_filters(limit=42)
        self.assertEqual(['limit=42'], result)

    def test_limit_0(self) -> None:
        result = utils.common_filters(limit=0)
        self.assertEqual([], result)

    def test_other(self) -> None:
        for key in ('marker', 'sort_key', 'sort_dir'):
            result = utils.common_filters(**{key: 'test'})
            self.assertEqual(['%s=test' % key], result)

    def test_fields(self) -> None:
        result = utils.common_filters(fields=['a', 'b', 'c'])
        self.assertEqual(['fields=a,b,c'], result)


@mock.patch.object(subprocess, 'Popen', autospec=True)
class MakeConfigDriveTest(test_utils.BaseTestCase):

    def setUp(self) -> None:
        super(MakeConfigDriveTest, self).setUp()
        # expected genisoimage cmd
        self.genisoimage_cmd = ['genisoimage', '-o', mock.ANY,
                                '-ldots', '-allow-lowercase',
                                '-allow-multidot', '-l',
                                '-publisher', 'ironicclient-configdrive 0.1',
                                '-quiet', '-J', '-r', '-V',
                                'config-2', mock.ANY]

        self.mkisofs_cmd = ['mkisofs', '-o', mock.ANY,
                            '-ldots', '-allow-lowercase',
                            '-allow-multidot', '-l',
                            '-publisher', 'ironicclient-configdrive 0.1',
                            '-quiet', '-J', '-r', '-V',
                            'config-2', mock.ANY]

        self.xorrisofs_cmd = ['xorrisofs', '-o', mock.ANY,
                              '-ldots', '-allow-lowercase',
                              '-allow-multidot', '-l',
                              '-publisher', 'ironicclient-configdrive 0.1',
                              '-quiet', '-J', '-r', '-V',
                              'config-2', mock.ANY]

    def test_make_configdrive(self,
                              mock_popen: mock.MagicMock,
                              ) -> None:
        fake_process = mock.Mock(returncode=0)
        fake_process.communicate.return_value = ('', '')
        mock_popen.return_value = fake_process

        with utils.tempdir() as dirname:
            utils.make_configdrive(dirname)

        mock_popen.assert_called_once_with(self.genisoimage_cmd,
                                           stderr=subprocess.PIPE,
                                           stdout=subprocess.PIPE)
        fake_process.communicate.assert_called_once_with()

    def test_make_configdrive_fallsback(self,
                                        mock_popen: mock.MagicMock,
                                        ) -> None:
        fake_process = mock.Mock(returncode=0)
        fake_process.communicate.return_value = ('', '')
        mock_popen.side_effect = iter([OSError('boom'),
                                       OSError('boom'),
                                       fake_process])
        with utils.tempdir() as dirname:
            utils.make_configdrive(dirname)
        mock_popen.assert_has_calls([
            mock.call(self.genisoimage_cmd, stderr=subprocess.PIPE,
                      stdout=subprocess.PIPE),
            mock.call(self.mkisofs_cmd, stderr=subprocess.PIPE,
                      stdout=subprocess.PIPE),
            mock.call(self.xorrisofs_cmd, stderr=subprocess.PIPE,
                      stdout=subprocess.PIPE)
        ])
        fake_process.communicate.assert_called_once_with()

    @mock.patch.object(os, 'access', autospec=True)
    def test_make_configdrive_non_readable_dir(
        self,
        mock_access: mock.MagicMock,
        mock_popen: mock.MagicMock,
    ) -> None:
        mock_access.return_value = False
        self.assertRaises(exc.CommandError, utils.make_configdrive, 'fake-dir')
        mock_access.assert_called_once_with('fake-dir', os.R_OK)
        self.assertFalse(mock_popen.called)

    @mock.patch.object(os, 'access', autospec=True)
    def test_make_configdrive_oserror(
        self,
        mock_access: mock.MagicMock,
        mock_popen: mock.MagicMock,
    ) -> None:
        mock_access.return_value = True
        mock_popen.side_effect = OSError('boom')

        self.assertRaises(exc.CommandError, utils.make_configdrive, 'fake-dir')
        mock_access.assert_called_once_with('fake-dir', os.R_OK)
        mock_popen.assert_has_calls([
            mock.call(self.genisoimage_cmd, stderr=subprocess.PIPE,
                      stdout=subprocess.PIPE),
            mock.call(self.mkisofs_cmd, stderr=subprocess.PIPE,
                      stdout=subprocess.PIPE),
            mock.call(self.xorrisofs_cmd, stderr=subprocess.PIPE,
                      stdout=subprocess.PIPE)
        ])

    @mock.patch.object(os, 'access', autospec=True)
    def test_make_configdrive_non_zero_returncode(
        self,
        mock_access: mock.MagicMock,
        mock_popen: mock.MagicMock,
    ) -> None:
        fake_process = mock.Mock(returncode=123)
        fake_process.communicate.return_value = ('', '')
        mock_popen.return_value = fake_process

        self.assertRaises(exc.CommandError, utils.make_configdrive, 'fake-dir')
        mock_access.assert_called_once_with('fake-dir', os.R_OK)
        mock_popen.assert_called_once_with(self.genisoimage_cmd,
                                           stderr=subprocess.PIPE,
                                           stdout=subprocess.PIPE)
        fake_process.communicate.assert_called_once_with()


class GetFromStdinTest(test_utils.BaseTestCase):

    @mock.patch.object(sys, 'stdin', autospec=True)
    def test_get_from_stdin(self,
                            mock_stdin: mock.MagicMock,
                            ) -> None:
        contents = '[{"step": "upgrade", "interface": "deploy"}]'
        mock_stdin.read.return_value = contents
        desc = 'something'

        info = utils.get_from_stdin(desc)
        self.assertEqual(info, contents)
        mock_stdin.read.assert_called_once_with()

    @mock.patch.object(sys, 'stdin', autospec=True)
    def test_get_from_stdin_fail(self,
                                 mock_stdin: mock.MagicMock,
                                 ) -> None:
        mock_stdin.read.side_effect = IOError
        desc = 'something'

        self.assertRaises(exc.InvalidAttribute, utils.get_from_stdin, desc)
        mock_stdin.read.assert_called_once_with()


class HandleJsonFileTest(test_utils.BaseTestCase):

    def test_handle_json_or_file_arg(self) -> None:
        cleansteps = '[{"step": "upgrade", "interface": "deploy"}]'
        steps = utils.handle_json_or_file_arg(cleansteps)
        self.assertEqual(json.loads(cleansteps), steps)

    def test_handle_json_or_file_arg_bad_json(self) -> None:
        cleansteps = '{foo invalid: json{'
        self.assertRaisesRegex(exc.InvalidAttribute,
                               'is not a file and cannot be parsed as JSON',
                               utils.handle_json_or_file_arg, cleansteps)

    def test_handle_json_or_file_arg_file(self) -> None:
        contents = '[{"step": "upgrade", "interface": "deploy"}]'

        with tempfile.NamedTemporaryFile(mode='w') as f:
            f.write(contents)
            f.flush()
            steps = utils.handle_json_or_file_arg(f.name)

        self.assertEqual(json.loads(contents), steps)

    def test_handle_yaml_or_file_arg_file(self) -> None:
        contents = '''---
- step: upgrade
  interface: deploy'''

        with tempfile.NamedTemporaryFile(mode='w') as f:
            f.write(contents)
            f.flush()
            steps = utils.handle_json_or_file_arg(f.name)

        self.assertEqual([{"step": "upgrade", "interface": "deploy"}], steps)

    @mock.patch.object(builtins, 'open', autospec=True)
    def test_handle_json_or_file_arg_file_fail(
        self,
        mock_open: mock.MagicMock,
    ) -> None:
        mock_open.return_value.__enter__.side_effect = IOError

        with tempfile.NamedTemporaryFile(mode='w') as f:
            self.assertRaisesRegex(exc.InvalidAttribute,
                                   "from file",
                                   utils.handle_json_or_file_arg, f.name)
            mock_open.assert_called_once_with(f.name, 'r')


class GetJsonDataTest(test_utils.BaseTestCase):

    def test_success(self) -> None:
        result = utils.get_json_data(b'\n{"answer": 42}')
        self.assertEqual({"answer": 42}, result)

    def test_definitely_not_json(self) -> None:
        self.assertIsNone(utils.get_json_data(b'0x010x020x03'))

    def test_could_be_json(self) -> None:
        self.assertIsNone(utils.get_json_data(b'{"hahaha, just kidding\x00'))



# Copyright 2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Base utilities to build API operation managers and objects on top of.
"""

from __future__ import annotations

import abc
import copy
from typing import Any, cast, Generic, overload, TypeVar
from urllib import parse as urlparse

from ironicclient.common.apiclient import base
from ironicclient.common.http import SessionClient
from ironicclient import exc


class Resource(base.Resource):
    """Represents a particular instance of an object (tenant, user, etc).

    This is pretty much just a bag for attributes.
    """

    def to_dict(self) -> dict[str, Any]:
        return copy.deepcopy(self._info)


ResourceT = TypeVar('ResourceT', bound=Resource)
ResourceAltT = TypeVar('ResourceAltT', bound=Resource)


def getid(obj: str | Resource) -> str:
    """Wrapper to get  object's ID.

    Abstracts the common pattern of allowing both an object or an
    object's ID (UUID) as a parameter when dealing with relationships.
    """
    if isinstance(obj, str):
        return obj
    try:
        return cast(str, obj.id)
    except AttributeError:
        return cast(str, obj)


class Manager(Generic[ResourceT], metaclass=abc.ABCMeta):
    """Provides  CRUD operations with a particular API."""

    def __init__(self, api: SessionClient) -> None:
        self.api = api
        self.client = api

    def _path(self, resource_id: str | None = None) -> str:
        """Returns a request path for a given resource identifier.

        :param resource_id: Identifier of the resource to generate the request
                            path.
        """
        return ('/v1/%s/%s' % (self._resource_name, resource_id)
                if resource_id else '/v1/%s' % self._resource_name)

    @property
    @abc.abstractmethod
    def resource_class(self) -> type[ResourceT]:
        """The resource class"""

    @property
    @abc.abstractmethod
    def _resource_name(self) -> str:
        """The resource name."""

    def _get(
        self,
        resource_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> ResourceT | None:
        """Retrieve a resource.

        :param resource_id: Identifier of the resource.
        :param fields: List of specific fields to be returned.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :raises exc.ValidationError: For invalid resource_id arg value.
        """

        if not resource_id:
            raise exc.ValidationError(
                "The identifier argument is invalid. "
                "Value provided: {!r}".format(resource_id))

        if fields is not None:
            resource_id = '%s?fields=' % resource_id
            resource_id += ','.join(fields)

        try:
            return self._list(
                self._path(resource_id),
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id)[0]
        except IndexError:
            return None

    def _get_as_dict(
        self,
        resource_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, Any]:
        """Retrieve a resource as a dictionary

        :param resource_id: Identifier of the resource.
        :param fields: List of specific fields to be returned.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :returns: a dictionary representing the resource; may be empty
        """

        resource = self._get(resource_id, fields=fields,
                             os_ironic_api_version=os_ironic_api_version,
                             global_request_id=global_request_id)
        if resource:
            return resource.to_dict()
        else:
            return {}

    def _format_body_data(
        self,
        body: dict[str, Any],
        response_key: str | None,
    ) -> list[Any]:
        if response_key:
            try:
                data = body[response_key]
            except KeyError:
                return []
        else:
            data = body

        if not isinstance(data, list):
            return [data]

        return data

    @overload
    def _list_pagination(
        self,
        url: str,
        response_key: str | None = None,
        obj_class: None = None,
        limit: int | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[ResourceT]:
        ...

    @overload
    def _list_pagination(
        self,
        url: str,
        response_key: str | None = None,
        obj_class: type[ResourceAltT] = ...,
        limit: int | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[ResourceAltT]:
        ...

    def _list_pagination(
        self,
        url: str,
        response_key: str | None = None,
        obj_class: type[Resource] | None = None,
        limit: int | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Any]:
        """Retrieve a list of items.

        The Ironic API is configured to return a maximum number of
        items per request, (see Ironic's api.max_limit option). This
        iterates over the 'next' link (pagination) in the responses,
        to get the number of items specified by 'limit'. If 'limit'
        is None this function will continue pagination until there are
        no more values to be returned.

        :param url: a partial URL, e.g. '/nodes'
        :param response_key: the key to be looked up in response
            dictionary, e.g. 'nodes'
        :param obj_class: class for constructing the returned objects.
        :param limit: maximum number of items to return. If None returns
            everything.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        if obj_class is None:
            obj_class = self.resource_class

        if limit is not None:
            limit = int(limit)
        headers: dict[str, str] = {}
        if os_ironic_api_version is not None:
            headers[
                'X-OpenStack-Ironic-API-Version'] = os_ironic_api_version
        if global_request_id is not None:
            headers["X-Openstack-Request-Id"] = global_request_id

        # NOTE(jroll)
        # endpoint_trimmed is what is prepended if we only pass a path
        # to json_request. This might be something like
        # https://example.com:4443/baremetal
        # The code below which parses the 'next' URL keeps any path prefix
        # we're using, so it ends up calling json_request() with e.g.
        # /baremetal/v1/nodes. When this is appended to endpoint_trimmed,
        # we end up with a full URL of e.g.
        # https://example.com:4443/baremetal/baremetal/v1/nodes.
        # Parse out the prefix from the base endpoint here, so we can strip
        # it below and make sure we're passing a correct path.
        endpoint_parts = urlparse.urlparse(self.api.endpoint_trimmed)
        url_path_prefix = endpoint_parts[2]

        object_list: list[Resource] = []
        object_count = 0
        limit_reached = False
        while url:
            resp, body = self.api.json_request(
                'GET', url, headers=headers)
            if not isinstance(body, dict):
                raise exc.InvalidAttribute(
                    'API response body must be a JSON object; got %s' %
                    type(body).__name__)
            data = self._format_body_data(body, response_key)
            for obj in data:
                item = obj_class(
                    self,
                    obj,
                    loaded=True,
                )
                object_list.append(item)
                object_count += 1
                if limit and object_count >= limit:
                    # break the for loop
                    limit_reached = True
                    break

            # break the while loop and return
            if limit_reached:
                break

            url = body.get('next', '')
            if url:
                # NOTE(lucasagomes): We need to edit the URL to remove
                # the scheme and netloc
                url_parts = list(urlparse.urlparse(url))
                url_parts[0] = url_parts[1] = ''
                if url_path_prefix:
                    url_parts[2] = url_parts[2].replace(
                        url_path_prefix, '', 1)
                url = urlparse.urlunparse(url_parts)

        return object_list

    def __list(
        self,
        url: str,
        response_key: str | None = None,
        body: dict[str, Any] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Any]:
        headers: dict[str, str] = {}

        if os_ironic_api_version is not None:
            headers[
                'X-OpenStack-Ironic-API-Version'] = os_ironic_api_version
        if global_request_id is not None:
            headers["X-Openstack-Request-Id"] = global_request_id

        _resp, json_body = self.api.json_request(
            'GET', url, headers=headers)
        if not isinstance(json_body, dict):
            raise exc.InvalidAttribute(
                'API response body must be a JSON object; got %s' %
                type(json_body).__name__)

        data = self._format_body_data(json_body, response_key)
        return data

    @overload
    def _list(
        self,
        url: str,
        response_key: str | None = None,
        obj_class: None = None,
        body: dict[str, Any] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[ResourceT]:
        ...

    @overload
    def _list(
        self,
        url: str,
        response_key: str | None = None,
        obj_class: type[ResourceAltT] = ...,
        body: dict[str, Any] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[ResourceAltT]:
        ...

    def _list(
        self,
        url: str,
        response_key: str | None = None,
        obj_class: type[Resource] | None = None,
        body: dict[str, Any] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Any]:
        if obj_class is None:
            obj_class = self.resource_class

        data = self.__list(
            url,
            response_key=response_key,
            body=body,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )
        return [
            obj_class(self, res, loaded=True)
            for res in data
            if res
        ]

    def _list_primitives(
        self,
        url: str,
        response_key: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Any]:
        return self.__list(
            url,
            response_key=response_key,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    @overload
    def _update(
        self,
        resource_id: str,
        patch: None,
        method: str = 'PATCH',
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> None:
        ...

    @overload
    def _update(
        self,
        resource_id: str,
        patch: list[dict[str, Any]],
        method: str = 'PATCH',
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> ResourceT:
        ...

    def _update(
        self,
        resource_id: str,
        patch: list[dict[str, Any]] | None,
        method: str = 'PATCH',
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> ResourceT | None:
        """Update a resource.

        :param resource_id: Resource identifier.
        :param patch: New version of a given resource, a dictionary or None.
        :param method: Name of the method for the request.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :param params: query parameters to pass.
        """

        url = self._path(resource_id)
        headers: dict[str, str] = {}
        if os_ironic_api_version is not None:
            headers[
                'X-OpenStack-Ironic-API-Version'] = os_ironic_api_version
        if global_request_id is not None:
            headers["X-Openstack-Request-Id"] = global_request_id
        _resp, body = self.api.json_request(
            method, url, body=patch, headers=headers, params=params)
        # PATCH/PUT requests may not return a body
        if body:
            return self.resource_class(
                self, cast(dict[str, Any], body),
            )
        return None

    def _delete(
        self,
        resource_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        """Delete a resource.

        :param resource_id: Resource identifier.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        headers: dict[str, str] = {}
        if os_ironic_api_version is not None:
            headers["X-OpenStack-Ironic-API-Version"] = os_ironic_api_version
        if global_request_id is not None:
            headers["X-Openstack-Request-Id"] = global_request_id
        self.api.raw_request('DELETE', self._path(resource_id),
                             headers=headers)


class CreateManager(Manager[ResourceT], metaclass=abc.ABCMeta):
    """Provides creation operations with a particular API."""

    @property
    @abc.abstractmethod
    def _creation_attributes(self) -> list[str]:
        """A list of required creation attributes for a resource type."""

    def create(
        self,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        **kwargs: Any,
    ) -> ResourceT | None:
        """Create a resource based on a kwargs dictionary of attributes.

        :param kwargs: A dictionary containing the attributes of the resource
                       that will be created.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :raises exc.InvalidAttribute: For invalid attributes that are not
                                      needed to create the resource.
        """

        new = {}
        invalid = []
        for (key, value) in kwargs.items():
            if key in self._creation_attributes:
                new[key] = value
            else:
                invalid.append(key)
        if invalid:
            raise exc.InvalidAttribute(
                'The attribute(s) "%(attrs)s" are invalid; they are not '
                'needed to create %(resource)s.' %
                {'resource': self._resource_name,
                 'attrs': '","'.join(invalid)})
        headers: dict[str, str] = {}
        if os_ironic_api_version is not None:
            headers['X-OpenStack-Ironic-API-Version'] = os_ironic_api_version
        if global_request_id is not None:
            headers["X-Openstack-Request-Id"] = global_request_id
        url = self._path()
        _resp, body = self.api.json_request(
            'POST', url, body=new, headers=headers,
        )
        if body:
            return self.resource_class(
                self, cast(dict[str, Any], body),
            )
        return None



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import annotations

import builtins
from unittest import mock

import jsonschema

from ironicclient import exc
from ironicclient.tests.unit import utils
from ironicclient.v1 import create_resources

valid_json = {
    "chassis": [{
        "description": "testing resources import",
        "nodes": [{
            "driver": "agent_ssh",
            "extra": {
                "kv1": True,
                "vk1": None
            },
            "properties": {
                "a": "c"
            },
            "ports": [{
                "address": "00:00:00:00:00:01",
                "extra": {
                    "a": "b"
                }
            }],
            "portgroups": [{
                "address": "00:00:00:00:00:02",
                "name": "portgroup1",
                "ports": [{
                    "address": "00:00:00:00:00:03"
                }]
            }]
        }]
    }],
    "nodes": [{
        "driver": "fake",
        "driver_info": {
            "fake_key": "fake",
            "dict_prop": {
                "a": "b"
            },
            "arr_prop": [
                1, 2, 3
            ]
        },
        "chassis_uuid": "10f99593-b8c2-4fcb-8858-494c1a47cee6"
    }]
}

ironic_pov_invalid_json = {
    "nodes": [{
        "driver": "non_existent",
        "ports": [{
            "address": "invalid_address"
        }]
    }]
}

schema_pov_invalid_json = {"meow": "woof!"}


class CreateSchemaTest(utils.BaseTestCase):

    def test_schema(self) -> None:
        schema = create_resources._CREATE_SCHEMA
        jsonschema.validate(valid_json, schema)
        jsonschema.validate(ironic_pov_invalid_json, schema)
        self.assertRaises(jsonschema.ValidationError, jsonschema.validate,
                          schema_pov_invalid_json, schema)


class CreateResourcesTest(utils.BaseTestCase):

    def setUp(self) -> None:
        super(CreateResourcesTest, self).setUp()
        self.client = mock.MagicMock()

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    @mock.patch.object(create_resources, 'create_chassis', autospec=True)
    @mock.patch.object(jsonschema, 'validate', autospec=True)
    @mock.patch.object(create_resources, 'load_from_file',
                       side_effect=[valid_json], autospec=True)
    def test_create_resources(
            self,
            mock_load: mock.MagicMock,
            mock_validate: mock.MagicMock,
            mock_chassis: mock.MagicMock,
            mock_nodes: mock.MagicMock,
    ) -> None:
        resources_files = ['file.json']
        create_resources.create_resources(self.client, resources_files)
        mock_load.assert_has_calls([
            mock.call('file.json')
        ])
        mock_validate.assert_called_once_with(valid_json, mock.ANY)
        mock_chassis.assert_called_once_with(self.client,
                                             valid_json['chassis'])
        mock_nodes.assert_called_once_with(self.client,
                                           valid_json['nodes'])

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    @mock.patch.object(create_resources, 'create_chassis', autospec=True)
    @mock.patch.object(jsonschema, 'validate', autospec=True)
    @mock.patch.object(create_resources, 'load_from_file',
                       side_effect=exc.ClientException, autospec=True)
    def test_create_resources_cannot_read_schema(
            self,
            mock_load: mock.MagicMock,
            mock_validate: mock.MagicMock,
            mock_chassis: mock.MagicMock,
            mock_nodes: mock.MagicMock,
    ) -> None:
        resources_files = ['file.json']
        self.assertRaises(exc.ClientException,
                          create_resources.create_resources,
                          self.client, resources_files)
        mock_load.assert_called_once_with('file.json')
        self.assertFalse(mock_validate.called)
        self.assertFalse(mock_chassis.called)
        self.assertFalse(mock_nodes.called)

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    @mock.patch.object(create_resources, 'create_chassis', autospec=True)
    @mock.patch.object(jsonschema, 'validate',
                       side_effect=jsonschema.ValidationError(''),
                       autospec=True)
    @mock.patch.object(create_resources, 'load_from_file',
                       side_effect=[schema_pov_invalid_json], autospec=True)
    def test_create_resources_validation_fails(
            self,
            mock_load: mock.MagicMock,
            mock_validate: mock.MagicMock,
            mock_chassis: mock.MagicMock,
            mock_nodes: mock.MagicMock,
    ) -> None:
        resources_files = ['file.json']
        self.assertRaises(exc.ClientException,
                          create_resources.create_resources,
                          self.client, resources_files)
        mock_load.assert_has_calls([
            mock.call('file.json')
        ])
        mock_validate.assert_called_once_with(schema_pov_invalid_json,
                                              mock.ANY)
        self.assertFalse(mock_chassis.called)
        self.assertFalse(mock_nodes.called)

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    @mock.patch.object(create_resources, 'create_chassis', autospec=True)
    @mock.patch.object(jsonschema, 'validate',
                       side_effect=[None, jsonschema.ValidationError('')],
                       autospec=True)
    @mock.patch.object(create_resources, 'load_from_file',
                       side_effect=[valid_json, schema_pov_invalid_json],
                       autospec=True)
    def test_create_resources_validation_fails_multiple(
            self,
            mock_load: mock.MagicMock,
            mock_validate: mock.MagicMock,
            mock_chassis: mock.MagicMock,
            mock_nodes: mock.MagicMock,
    ) -> None:
        resources_files = ['file.json', 'file2.json']
        self.assertRaises(exc.ClientException,
                          create_resources.create_resources,
                          self.client, resources_files)
        mock_load.assert_has_calls([
            mock.call('file.json'), mock.call('file2.json')
        ])
        mock_validate.assert_has_calls([
            mock.call(valid_json, mock.ANY),
            mock.call(schema_pov_invalid_json, mock.ANY)
        ])
        self.assertFalse(mock_chassis.called)
        self.assertFalse(mock_nodes.called)

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    @mock.patch.object(create_resources, 'create_chassis', autospec=True)
    @mock.patch.object(jsonschema, 'validate', autospec=True)
    @mock.patch.object(create_resources, 'load_from_file',
                       side_effect=[ironic_pov_invalid_json], autospec=True)
    def test_create_resources_ironic_fails_to_create(
            self,
            mock_load: mock.MagicMock,
            mock_validate: mock.MagicMock,
            mock_chassis: mock.MagicMock,
            mock_nodes: mock.MagicMock,
    ) -> None:
        mock_nodes.return_value = [exc.ClientException('cannot create that')]
        mock_chassis.return_value = []
        resources_files = ['file.json']
        self.assertRaises(exc.ClientException,
                          create_resources.create_resources,
                          self.client, resources_files)
        mock_load.assert_has_calls([
            mock.call('file.json')
        ])
        mock_validate.assert_called_once_with(ironic_pov_invalid_json,
                                              mock.ANY)
        mock_chassis.assert_called_once_with(self.client, [])
        mock_nodes.assert_called_once_with(
            self.client, ironic_pov_invalid_json['nodes'])


class LoadFromFileTest(utils.BaseTestCase):

    @mock.patch.object(builtins, 'open',
                       mock.mock_open(read_data='{"a": "b"}'))
    def test_load_json(self) -> None:
        fname = 'abc.json'
        res = create_resources.load_from_file(fname)
        self.assertEqual({'a': 'b'}, res)

    @mock.patch.object(builtins, 'open',
                       mock.mock_open(read_data='{"a": "b"}'))
    def test_load_unknown_extension(self) -> None:
        fname = 'abc'
        self.assertRaisesRegex(exc.ClientException,
                               'must have .json or .yaml extension',
                               create_resources.load_from_file, fname)

    @mock.patch.object(builtins, 'open', autospec=True)
    def test_load_ioerror(self, mock_open: mock.MagicMock) -> None:
        mock_open.side_effect = IOError('file does not exist')
        fname = 'abc.json'
        self.assertRaisesRegex(exc.ClientException,
                               'Cannot read file',
                               create_resources.load_from_file, fname)

    @mock.patch.object(builtins, 'open',
                       mock.mock_open(read_data='{{bbb'))
    def test_load_incorrect_json(self) -> None:
        fname = 'abc.json'
        self.assertRaisesRegex(
            exc.ClientException, 'File "%s" is invalid' % fname,
            create_resources.load_from_file, fname)

    @mock.patch.object(builtins, 'open',
                       mock.mock_open(read_data='---\na: b'))
    def test_load_yaml(self) -> None:
        fname = 'abc.yaml'
        res = create_resources.load_from_file(fname)
        self.assertEqual({'a': 'b'}, res)

    @mock.patch.object(builtins, 'open',
                       mock.mock_open(read_data='---\na-: - b'))
    def test_load_incorrect_yaml(self) -> None:
        fname = 'abc.yaml'
        self.assertRaisesRegex(
            exc.ClientException, 'File "%s" is invalid' % fname,
            create_resources.load_from_file, fname)


class CreateMethodsTest(utils.BaseTestCase):

    def setUp(self) -> None:
        super(CreateMethodsTest, self).setUp()
        self.client = mock.MagicMock()

    def test_create_single_node(self) -> None:
        params = {'driver': 'fake'}
        self.client.node.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual(
            ('uuid', None),
            create_resources.create_single_node(self.client, **params)
        )
        self.client.node.create.assert_called_once_with(driver='fake')

    def test_create_single_node_with_traits(self) -> None:
        params = {'driver': 'fake', 'traits': ['CUSTOM_PERFORMANCE']}
        self.client.node.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual(
            ('uuid', None),
            create_resources.create_single_node(self.client, **params)
        )
        self.client.node.create.assert_called_once_with(driver='fake')
        self.client.node.set_traits.assert_called_once_with(
            mock.ANY, ['CUSTOM_PERFORMANCE'])

    def test_create_single_node_with_ports(self) -> None:
        params = {'driver': 'fake', 'ports': ['some ports here']}
        self.client.node.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual(
            ('uuid', None),
            create_resources.create_single_node(self.client, **params)
        )
        self.client.node.create.assert_called_once_with(driver='fake')

    def test_create_single_node_with_portgroups(self) -> None:
        params = {'driver': 'fake', 'portgroups': ['some portgroups']}
        self.client.node.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual(
            ('uuid', None),
            create_resources.create_single_node(self.client, **params)
        )
        self.client.node.create.assert_called_once_with(driver='fake')

    def test_create_single_node_raises_client_exception(self) -> None:
        params = {'driver': 'fake'}
        e = exc.ClientException('foo')
        self.client.node.create.side_effect = e
        res, err = create_resources.create_single_node(self.client, **params)
        self.assertIsNone(res)
        self.assertIsInstance(err, exc.ClientException)
        self.assertIn('Unable to create the node', str(err))
        self.client.node.create.assert_called_once_with(driver='fake')

    def test_create_single_node_raises_invalid_exception(self) -> None:
        params = {'driver': 'fake'}
        e = exc.InvalidAttribute('foo')
        self.client.node.create.side_effect = e
        res, err = create_resources.create_single_node(self.client, **params)
        self.assertIsNone(res)
        self.assertIsInstance(err, exc.InvalidAttribute)
        self.assertIn('Cannot create the node with attributes', str(err))
        self.client.node.create.assert_called_once_with(driver='fake')

    def test_create_single_port(self) -> None:
        params = {'address': 'fake-address', 'node_uuid': 'fake-node-uuid'}
        self.client.port.create.return_value = mock.Mock(uuid='fake-port-uuid')
        self.assertEqual(
            ('fake-port-uuid', None),
            create_resources.create_single_port(self.client, **params)
        )
        self.client.port.create.assert_called_once_with(**params)

    def test_create_single_portgroup(self) -> None:
        params = {'address': 'fake-address', 'node_uuid': 'fake-node-uuid'}
        self.client.portgroup.create.return_value = mock.Mock(
            uuid='fake-portgroup-uuid')
        self.assertEqual(
            ('fake-portgroup-uuid', None),
            create_resources.create_single_portgroup(self.client, **params)
        )
        self.client.portgroup.create.assert_called_once_with(**params)

    def test_create_single_portgroup_with_ports(self) -> None:
        params = {'ports': ['some ports'], 'node_uuid': 'fake-node-uuid'}
        self.client.portgroup.create.return_value = mock.Mock(
            uuid='fake-portgroup-uuid')
        self.assertEqual(
            ('fake-portgroup-uuid', None),
            create_resources.create_single_portgroup(
                self.client, **params)
        )
        self.client.portgroup.create.assert_called_once_with(
            node_uuid='fake-node-uuid')

    def test_create_single_chassis(self) -> None:
        self.client.chassis.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual(
            ('uuid', None),
            create_resources.create_single_chassis(self.client)
        )
        self.client.chassis.create.assert_called_once_with()

    def test_create_single_chassis_with_nodes(self) -> None:
        params = {'nodes': ['some nodes here']}
        self.client.chassis.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual(
            ('uuid', None),
            create_resources.create_single_chassis(self.client, **params)
        )
        self.client.chassis.create.assert_called_once_with()

    def test_create_ports(self) -> None:
        port: dict[str, object] = {'address': 'fake-address'}
        port_with_node_uuid = port.copy()
        port_with_node_uuid.update(node_uuid='fake-node-uuid')
        self.client.port.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_ports(self.client, [port],
                                                           'fake-node-uuid'))
        self.client.port.create.assert_called_once_with(**port_with_node_uuid)

    def test_create_ports_two_node_uuids(self) -> None:
        port: dict[str, object] = {'address': 'fake-address',
                                   'node_uuid': 'node-uuid-1'}
        errs = create_resources.create_ports(self.client, [port],
                                             'node-uuid-2')
        self.assertIsInstance(errs[0], exc.ClientException)
        self.assertEqual(1, len(errs))
        self.assertFalse(self.client.port.create.called)

    def test_create_ports_two_portgroup_uuids(self) -> None:
        port: dict[str, object] = {'address': 'fake-address',
                                   'node_uuid': 'node-uuid-1',
                                   'portgroup_uuid': 'pg-uuid-1'}
        errs = create_resources.create_ports(self.client, [port],
                                             'node-uuid-1', 'pg-uuid-2')
        self.assertEqual(1, len(errs))
        self.assertIsInstance(errs[0], exc.ClientException)
        self.assertIn('port group', str(errs[0]))
        self.assertFalse(self.client.port.create.called)

    @mock.patch.object(create_resources, 'create_portgroups', autospec=True)
    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_nodes(
            self,
            mock_create_ports: mock.MagicMock,
            mock_create_portgroups: mock.MagicMock,
    ) -> None:
        node: dict[str, object] = {'driver': 'fake',
                                   'ports': ['list of ports'],
                                   'portgroups': ['list of portgroups']}
        self.client.node.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_nodes(self.client,
                                                           [node]))
        self.client.node.create.assert_called_once_with(driver='fake')
        mock_create_ports.assert_called_once_with(
            self.client, ['list of ports'], node_uuid='uuid')
        mock_create_portgroups.assert_called_once_with(
            self.client, ['list of portgroups'], node_uuid='uuid')

    @mock.patch.object(create_resources, 'create_portgroups', autospec=True)
    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_nodes_exception(
            self,
            mock_create_ports: mock.MagicMock,
            mock_create_portgroups: mock.MagicMock,
    ) -> None:
        node: dict[str, object] = {'driver': 'fake',
                                   'ports': ['list of ports'],
                                   'portgroups': ['list of portgroups']}
        self.client.node.create.side_effect = exc.ClientException('bar')
        errs = create_resources.create_nodes(self.client, [node])
        self.assertIsInstance(errs[0], exc.ClientException)
        self.assertEqual(1, len(errs))
        self.client.node.create.assert_called_once_with(driver='fake')
        self.assertFalse(mock_create_ports.called)
        self.assertFalse(mock_create_portgroups.called)

    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_nodes_two_chassis_uuids(
            self,
            mock_create_ports: mock.MagicMock,
    ) -> None:
        node: dict[str, object] = {'driver': 'fake',
                                   'ports': ['list of ports'],
                                   'chassis_uuid': 'chassis-uuid-1'}
        errs = create_resources.create_nodes(self.client, [node],
                                             chassis_uuid='chassis-uuid-2')
        self.assertFalse(self.client.node.create.called)
        self.assertFalse(mock_create_ports.called)
        self.assertEqual(1, len(errs))
        self.assertIsInstance(errs[0], exc.ClientException)

    @mock.patch.object(create_resources, 'create_portgroups', autospec=True)
    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_nodes_no_ports_portgroups(
            self,
            mock_create_ports: mock.MagicMock,
            mock_create_portgroups: mock.MagicMock,
    ) -> None:
        node: dict[str, object] = {'driver': 'fake'}
        self.client.node.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_nodes(self.client,
                                                           [node]))
        self.client.node.create.assert_called_once_with(driver='fake')
        self.assertFalse(mock_create_ports.called)
        self.assertFalse(mock_create_portgroups.called)

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    def test_create_chassis(self, mock_create_nodes: mock.MagicMock) -> None:
        chassis: dict[str, object] = {'description': 'fake',
                                      'nodes': ['list of nodes']}
        self.client.chassis.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_chassis(self.client,
                                                             [chassis]))
        self.client.chassis.create.assert_called_once_with(description='fake')
        mock_create_nodes.assert_called_once_with(
            self.client, ['list of nodes'], chassis_uuid='uuid')

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    def test_create_chassis_exception(
            self,
            mock_create_nodes: mock.MagicMock,
    ) -> None:
        chassis: dict[str, object] = {'description': 'fake',
                                      'nodes': ['list of nodes']}
        self.client.chassis.create.side_effect = exc.ClientException('bar')
        errs = create_resources.create_chassis(self.client, [chassis])
        self.client.chassis.create.assert_called_once_with(description='fake')
        self.assertFalse(mock_create_nodes.called)
        self.assertEqual(1, len(errs))
        self.assertIsInstance(errs[0], exc.ClientException)

    @mock.patch.object(create_resources, 'create_nodes', autospec=True)
    def test_create_chassis_no_nodes(
            self,
            mock_create_nodes: mock.MagicMock,
    ) -> None:
        chassis: dict[str, object] = {'description': 'fake'}
        self.client.chassis.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_chassis(self.client,
                                                             [chassis]))
        self.client.chassis.create.assert_called_once_with(description='fake')
        self.assertFalse(mock_create_nodes.called)

    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_portgroups(
            self,
            mock_create_ports: mock.MagicMock,
    ) -> None:
        portgroup: dict[str, object] = {'name': 'fake',
                                        'ports': ['list of ports']}
        portgroup_posted = {'name': 'fake', 'node_uuid': 'fake-node-uuid'}
        self.client.portgroup.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_portgroups(
            self.client, [portgroup], node_uuid='fake-node-uuid'))
        self.client.portgroup.create.assert_called_once_with(
            **portgroup_posted)
        mock_create_ports.assert_called_once_with(
            self.client, ['list of ports'], node_uuid='fake-node-uuid',
            portgroup_uuid='uuid')

    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_portgroups_exception(
            self,
            mock_create_ports: mock.MagicMock,
    ) -> None:
        portgroup: dict[str, object] = {'name': 'fake',
                                        'ports': ['list of ports']}
        portgroup_posted = {'name': 'fake', 'node_uuid': 'fake-node-uuid'}
        self.client.portgroup.create.side_effect = exc.ClientException('bar')
        errs = create_resources.create_portgroups(
            self.client, [portgroup], node_uuid='fake-node-uuid')
        self.client.portgroup.create.assert_called_once_with(
            **portgroup_posted)
        self.assertFalse(mock_create_ports.called)
        self.assertEqual(1, len(errs))
        self.assertIsInstance(errs[0], exc.ClientException)

    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_portgroups_two_node_uuids(
            self,
            mock_create_ports: mock.MagicMock,
    ) -> None:
        portgroup: dict[str, object] = {'name': 'fake',
                                        'node_uuid': 'fake-node-uuid-1',
                                        'ports': ['list of ports']}
        self.client.portgroup.create.side_effect = exc.ClientException('bar')
        errs = create_resources.create_portgroups(
            self.client, [portgroup], node_uuid='fake-node-uuid-2')
        self.assertFalse(self.client.portgroup.create.called)
        self.assertFalse(mock_create_ports.called)
        self.assertEqual(1, len(errs))
        self.assertIsInstance(errs[0], exc.ClientException)

    @mock.patch.object(create_resources, 'create_ports', autospec=True)
    def test_create_portgroups_no_ports(
            self,
            mock_create_ports: mock.MagicMock,
    ) -> None:
        portgroup: dict[str, object] = {'name': 'fake'}
        portgroup_posted = {'name': 'fake', 'node_uuid': 'fake-node-uuid'}
        self.client.portgroup.create.return_value = mock.Mock(uuid='uuid')
        self.assertEqual([], create_resources.create_portgroups(
            self.client, [portgroup], node_uuid='fake-node-uuid'))
        self.client.portgroup.create.assert_called_once_with(
            **portgroup_posted)
        self.assertFalse(mock_create_ports.called)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

import copy

from osc_lib.tests import utils as oscutils

from ironicclient.osc.v1 import baremetal_conductor
from ironicclient.tests.unit.osc.v1 import fakes as baremetal_fakes


class TestBaremetalConductor(baremetal_fakes.TestBaremetal):

    def setUp(self) -> None:
        super(TestBaremetalConductor, self).setUp()

        # Get a shortcut to the baremetal manager mock
        self.baremetal_mock = self.app.client_manager.baremetal
        self.baremetal_mock.reset_mock()


class TestBaremetalConductorList(TestBaremetalConductor):

    def setUp(self) -> None:
        super(TestBaremetalConductorList, self).setUp()

        self.baremetal_mock.conductor.list.return_value = [
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.CONDUCTOR),
                loaded=True,
            ),
        ]

        # Get the command object to test
        self.cmd = baremetal_conductor.ListBaremetalConductor(self.app, None)

    def test_conductor_list_no_options(self) -> None:
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': None,
            'limit': None,
        }

        self.baremetal_mock.conductor.list.assert_called_with(
            **kwargs
        )

        collist = (
            "hostname",
            "conductor_group",
            "alive",
        )
        self.assertEqual(collist, columns)
        datalist = ((
            baremetal_fakes.baremetal_hostname,
            baremetal_fakes.baremetal_conductor_group,
            baremetal_fakes.baremetal_alive,
        ), )
        self.assertEqual(datalist, tuple(data))

    def test_conductor_list_long(self) -> None:
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'detail': True,
            'marker': None,
            'limit': None,
        }

        self.baremetal_mock.conductor.list.assert_called_with(
            **kwargs
        )

        collist = tuple([
            'hostname',
            'conductor_group',
            'alive',
            'drivers',
            'created_at',
            'updated_at',
        ])
        self.assertEqual(collist, columns)

        fake_values = {
            'hostname': baremetal_fakes.baremetal_hostname,
            'conductor_group': baremetal_fakes.baremetal_conductor_group,
            'alive': baremetal_fakes.baremetal_alive,
            'drivers': baremetal_fakes.baremetal_drivers,
        }
        values = tuple(fake_values.get(name, '') for name in collist)
        self.assertEqual((values,), tuple(data))

    def test_conductor_list_fields(self) -> None:
        arglist = [
            '--fields', 'hostname', 'alive',
        ]
        verifylist = [
            ('fields', [['hostname', 'alive']]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('hostname', 'alive'),
        }

        self.baremetal_mock.conductor.list.assert_called_with(
            **kwargs
        )

    def test_conductor_list_fields_multiple(self) -> None:
        arglist = [
            '--fields', 'hostname', 'alive',
            '--fields', 'conductor_group',
        ]
        verifylist = [
            ('fields', [['hostname', 'alive'], ['conductor_group']])
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'marker': None,
            'limit': None,
            'detail': False,
            'fields': ('hostname', 'alive', 'conductor_group')
        }

        self.baremetal_mock.conductor.list.assert_called_with(
            **kwargs
        )

    def test_conductor_list_invalid_fields(self) -> None:
        arglist = [
            '--fields', 'hostname', 'invalid'
        ]
        verifylist = [
            ('fields', [['hostname', 'invalid']])
        ]

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)


class TestBaremetalConductorShow(TestBaremetalConductor):
    def setUp(self) -> None:
        super(TestBaremetalConductorShow, self).setUp()

        self.baremetal_mock.conductor.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(
                None,
                copy.deepcopy(baremetal_fakes.CONDUCTOR),
                loaded=True,
            ))

        # Get the command object to test
        self.cmd = baremetal_conductor.ShowBaremetalConductor(self.app, None)

    def test_conductor_show(self) -> None:
        arglist = ['xxxx.xxxx']
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        args = ['xxxx.xxxx']

        self.baremetal_mock.conductor.get.assert_called_with(
            *args, fields=None
        )

        collist = (
            'hostname',
            'conductor_group',
            'alive',
            'drivers',
        )
        self.assertEqual(collist, columns)
        datalist = (
            baremetal_fakes.baremetal_hostname,
            baremetal_fakes.baremetal_conductor_group,
            baremetal_fakes.baremetal_alive,
            baremetal_fakes.baremetal_drivers,
        )
        self.assertEqual(datalist, tuple(data))

    def test_conductor_show_no_conductor(self) -> None:
        arglist = []
        verifylist = []

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)

    def test_conductor_show_fields(self) -> None:
        arglist = [
            'xxxxx',
            '--fields', 'hostname', 'alive',
        ]
        verifylist = [
            ('conductor', 'xxxxx'),
            ('fields', [['hostname', 'alive']]),
        ]

        fake_cond = copy.deepcopy(baremetal_fakes.CONDUCTOR)
        fake_cond.pop('conductor_group')
        fake_cond.pop('drivers')
        self.baremetal_mock.conductor.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(None, fake_cond,
                                                  loaded=True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # DisplayCommandBase.take_action() returns two tuples
        columns, data = self.cmd.take_action(parsed_args)
        self.assertNotIn('conductor_group', columns)

        # Set expected values
        args = ['xxxxx']
        fields = ['hostname', 'alive']

        self.baremetal_mock.conductor.get.assert_called_with(
            *args, fields=fields
        )

    def test_conductor_show_fields_multiple(self) -> None:
        arglist = [
            'xxxxx',
            '--fields', 'hostname', 'alive',
            '--fields', 'conductor_group',
        ]
        verifylist = [
            ('conductor', 'xxxxx'),
            ('fields', [['hostname', 'alive'], ['conductor_group']])
        ]

        fake_cond = copy.deepcopy(baremetal_fakes.CONDUCTOR)
        fake_cond.pop('drivers')
        self.baremetal_mock.conductor.get.return_value = (
            baremetal_fakes.FakeBaremetalResource(None, fake_cond,
                                                  loaded=True))

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)
        self.assertNotIn('drivers', columns)
        # Set expected values
        args = ['xxxxx']
        fields = ['hostname', 'alive', 'conductor_group']

        self.baremetal_mock.conductor.get.assert_called_with(
            *args, fields=fields
        )

    def test_conductor_show_invalid_fields(self) -> None:
        arglist = [
            'xxxxx',
            '--fields', 'hostname', 'invalid'
        ]
        verifylist = [
            ('conductor', 'xxxxx'),
            ('fields', [['hostname', 'invalid']])
        ]

        self.assertRaises(oscutils.ParserException,
                          self.check_parser,
                          self.cmd, arglist, verifylist)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import annotations

import argparse
import logging

from ironicclient.common.i18n import _
from ironicclient.osc import command
from ironicclient.v1 import create_resources


class CreateBaremetal(command.Command):
    """Create resources from files"""

    log: logging.Logger = logging.getLogger(
        __name__ + ".CreateBaremetal")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser: argparse.ArgumentParser = super().get_parser(prog_name)

        parser.add_argument(
            "resource_files", metavar="<file>", nargs="+",
            help=_("File (.yaml or .json) containing descriptions of the "
                   "resources to create. Can be specified multiple times."))
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        create_resources.create_resources(
            self.app.client_manager.baremetal,
            parsed_args.resource_files)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from __future__ import annotations

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc


class Conductor(base.Resource):
    def __repr__(self) -> str:
        return "<Conductor %s>" % self._info


class ConductorManager(base.Manager[Conductor]):
    resource_class: type[Conductor] = Conductor
    _resource_name: str = 'conductors'

    def get(
        self,
        hostname: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        return self._get(
            resource_id=hostname,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list(
        self,
        marker: str | None = None,
        limit: int | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        fields: list[str] | None = None,
        detail: bool = False,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Conductor]:
        """Retrieve a list of conductors.

        :param marker: Optional, the hostname of a conductor, eg the last
                       conductor from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of ports to return.
            2) limit == 0, return the entire list of ports.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param detail: Optional, boolean whether to return detailed information
                       about conductors.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of conductors.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir,
            fields, detail)
        path = ''
        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path),
                "conductors",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "conductors",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from __future__ import annotations

from collections.abc import Sequence

from ironicclient.tests.unit import utils

module_str: str = 'ironicclient'


class ImportTest(utils.BaseTestCase):

    def check_exported_symbols(
        self, exported_symbols: Sequence[str],
    ) -> None:
        self.assertIn('client', exported_symbols)
        self.assertIn('exc', exported_symbols)
        self.assertIn('exceptions', exported_symbols)

    def test_import_objects(self) -> None:
        module = __import__(module_str)
        exported_symbols = dir(module)
        self.check_exported_symbols(exported_symbols)

    def test_default_import(self) -> None:
        default_imports = __import__(module_str, globals(), locals(), ['*'])
        exported_symbols = dir(default_imports)
        self.check_exported_symbols(exported_symbols)

    def test_import__all__(self) -> None:
        module = __import__(module_str)
        self.check_exported_symbols(module.__all__)



# Copyright 2013 Red Hat, Inc.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from typing import Any, cast

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient import exc


class Driver(base.Resource):
    def __repr__(self) -> str:
        return "<Driver %s>" % self._info


class DriverManager(base.Manager[Driver]):
    resource_class: type[Driver] = Driver
    _resource_name: str = 'drivers'

    def get(
        self,
        driver_name: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        fields: list[str] | None = None,
    ) -> base.Resource | None:
        return self._get(
            resource_id=driver_name,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def update(
        self,
        driver_name: str,
        patch: list[dict[str, Any]],
        http_method: str = 'PATCH',
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource:
        return self._update(
            resource_id=driver_name,
            patch=patch,
            method=http_method,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def delete(
        self,
        driver_name: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        return self._delete(
            resource_id=driver_name,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def properties(
        self,
        driver_name: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, str]:
        return cast(dict[str, str], self._get_as_dict(
            '%s/properties' % driver_name,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        ))

    def raid_logical_disk_properties(
        self,
        driver_name: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, str]:
        """Returns the RAID logical disk properties for the driver.

        :param driver_name: Name of the driver.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :returns: A dictionary containing the properties that can be mentioned
            for RAID logical disks and a textual description for them. It
            returns an empty dictionary on error.
        """
        info = None
        try:
            info = self._list(
                '/v1/drivers/%s/raid/logical_disk_properties'
                % driver_name,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id)[0]
        except IndexError:
            pass

        if info:
            return cast(dict[str, str], info.to_dict())
        return {}

    def vendor_passthru(
        self,
        driver_name: str,
        method: str,
        args: dict[str, Any] | None = None,
        http_method: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        """Issue requests for vendor-specific actions on a given driver.

        :param driver_name: The name of the driver.
        :param method: Name of the vendor method.
        :param args: Optional. The arguments to be passed to the method.
        :param http_method: The HTTP method to use on the request.
                            Defaults to POST.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        """
        if args is None:
            args = {}

        if http_method is None:
            http_method = 'POST'

        http_method = http_method.upper()

        path = "%s/vendor_passthru/%s" % (driver_name, method)
        if http_method in ('POST', 'PUT', 'PATCH'):
            return self.update(
                path,
                args,  # type: ignore[arg-type]
                http_method=http_method,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        elif http_method == 'DELETE':
            self.delete(
                path,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
            return None
        elif http_method == 'GET':
            return self.get(
                path,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            raise exc.InvalidAttribute(
                _('Unknown HTTP method: %s') % http_method)

    def get_vendor_passthru_methods(
        self,
        driver_name: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> dict[str, Any]:
        return self._get_as_dict(
            "%s/vendor_passthru/methods" % driver_name,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list(
        self,
        driver_type: str | None = None,
        detail: bool | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
        fields: list[str] | None = None,
    ) -> list[Driver]:
        """Retrieve a list of drivers.

        :param driver_type: Optional, string to filter the drivers by type.
                            Value should be 'classic' or 'dynamic'.
        :param detail: Optional, flag whether to return detailed information
                       about drivers. Default is None means not to send the arg
                       to the server due to older versions of the server cannot
                       handle filtering on detail.
        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.
        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.
        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.
        :returns: A list of drivers.
        """

        filters: list[str] = []
        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))
        if driver_type is not None:
            filters.append('type=%s' % driver_type)
        if detail is not None:
            filters.append('detail=%s' % detail)
        if fields is not None:
            filters.append('fields=%s' % ','.join(fields))

        path = ''
        if filters:
            path = '?' + '&'.join(filters)
        return self._list(
            self._path(path),
            self._resource_name,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )



# -*- coding: utf-8 -*-
#
# Copyright © 2013 Red Hat, Inc
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import annotations

from typing import Any

from ironicclient.common import base
from ironicclient.common.i18n import _
from ironicclient.common import utils
from ironicclient import exc


class Chassis(base.Resource):
    def __repr__(self) -> str:
        return "<Chassis %s>" % self._info


class ChassisManager(base.CreateManager[Chassis]):
    resource_class: type[Chassis] = Chassis
    _resource_name: str = 'chassis'
    _creation_attributes: list[str] = [
        'description', 'extra', 'uuid',
    ]

    def get(
        self,
        chassis_id: str,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource | None:
        return self._get(
            resource_id=chassis_id,
            fields=fields,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def delete(
        self,
        chassis_id: str,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> None:
        return self._delete(
            resource_id=chassis_id,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def update(
        self,
        chassis_id: str,
        patch: list[dict[str, Any]],
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> base.Resource:
        return self._update(
            resource_id=chassis_id,
            patch=patch,
            os_ironic_api_version=os_ironic_api_version,
            global_request_id=global_request_id,
        )

    def list_nodes(
        self,
        chassis_id: str,
        marker: str | None = None,
        limit: int | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        associated: bool | str | None = None,
        maintenance: bool | str | None = None,
        provision_state: str | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Chassis]:
        """List all the nodes for a given chassis.

        :param chassis_id: The UUID of the chassis.
        :param marker: Optional, the UUID of a node, eg the last
                       node from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of nodes to return.
            2) limit == 0, return the entire list of nodes.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about nodes.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param associated: Optional. Either a Boolean or a string
                           representation of a Boolean that indicates whether
                           to return a list of associated (True or "True") or
                           unassociated (False or "False") nodes.

        :param maintenance: Optional. Either a Boolean or a string
                            representation of a Boolean that indicates whether
                            to return nodes in maintenance mode (True or
                            "True"), or not in maintenance mode (False or
                            "False").

        :param provision_state: Optional. String value to get only nodes in
                                that provision state.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of nodes.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir, fields)

        if associated is not None:
            filters.append('associated=%s' % associated)
        if maintenance is not None:
            filters.append('maintenance=%s' % maintenance)
        if provision_state is not None:
            filters.append('provision_state=%s' % provision_state)

        path = "%s/nodes" % chassis_id
        if detail:
            path += '/detail'

        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path),
                "nodes",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "nodes",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )

    def list(
        self,
        marker: str | None = None,
        limit: int | None = None,
        sort_key: str | None = None,
        sort_dir: str | None = None,
        detail: bool = False,
        fields: list[str] | None = None,
        os_ironic_api_version: str | None = None,
        global_request_id: str | None = None,
    ) -> list[Chassis]:
        """Retrieve a list of chassis.

        :param marker: Optional, the UUID of a chassis, eg the last
                       chassis from a previous result set. Return
                       the next result set.
        :param limit: The maximum number of results to return per
                      request, if:

            1) limit > 0, the maximum number of chassis to return.
            2) limit == 0, return the entire list of chassis.
            3) limit param is NOT specified (None), the number of items
               returned respect the maximum imposed by the Ironic API
               (see Ironic's api.max_limit option).

        :param sort_key: Optional, field used for sorting.

        :param sort_dir: Optional, direction of sorting, either 'asc' (the
                         default) or 'desc'.

        :param detail: Optional, boolean whether to return detailed information
                       about chassis.

        :param fields: Optional, a list with a specified set of fields
                       of the resource to be returned. Can not be used
                       when 'detail' is set.

        :param os_ironic_api_version: String version (e.g. "1.35") to use for
            the request.  If not specified, the client's default is used.

        :param global_request_id: String containing global request ID header
            value (in form "req-<UUID>") to use for the request.

        :returns: A list of chassis.

        """
        if limit is not None:
            limit = int(limit)

        if detail and fields:
            raise exc.InvalidAttribute(
                _("Can't fetch a subset of fields "
                  "with 'detail' set"))

        filters = utils.common_filters(
            marker, limit, sort_key, sort_dir, fields)

        path = ''
        if detail:
            path += 'detail'
        if filters:
            path += '?' + '&'.join(filters)
        if limit is None:
            return self._list(
                self._path(path),
                "chassis",
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
        else:
            return self._list_pagination(
                self._path(path),
                "chassis",
                limit=limit,
                os_ironic_api_version=os_ironic_api_version,
                global_request_id=global_request_id,
            )
